# Customization of GPI component object assignment: Component Filters
#
# Example of usage:
#
#  j.f = 'myfile'  <=>  j.f = File('myfile')
#  j.application = 'DaVinci' <=> DaVinci()
#
#  j = Job()
#  j2 = Job(j)  <=> copy constructor
#
#
#
# Semantics:
#  gpi_proxy.x = v  --> f = select_filter(category_of(x)); f(v,schema_item_of(x))
#  gpi_proxy.y = [v1,v2] --> f = select_filter(category_of(x)); [f(v,schema_item_of(y)) for v in [v1,v2]]
#  x = X(y,...) --> f = select_filter(X._category); f(y,None)
#
# Component filters are applied *only* to component properties and they
# are applied *before* attribute filters.

# Component filter has the following signature:
#     filter(val,item) --> return a GangaObject instance or None if no conversion performed
#
#     item is None if the filter is called outside of the attribute assignment context
#
# If conversion takes place filter MUST return an object which is an
# instance (derived) of GangaObject.

# Void filter does nothing. This is the default filter if no other default
# has been defined.

from Ganga.GPIDev.Base.Proxy import isType, getName
from Ganga.Utility.Config import getConfig
from Ganga.Utility.Config.Config import ConfigError

# test configuration properties
config = getConfig('GPIComponentFilters')

def void_filter(val, item):
    return None


class _ComponentFilterManager(object):

    def __init__(self):
        # for each category there may be multiple filters registered, the one used being defined
        # in the configuration file in [GPIComponentFilters]
        # e.g: {'datasets':{'lhcbdatasets':lhcbFilter,
        # 'testdatasets':testFilter}...}
        self._dict = {}
        self.default = None

    def __setitem__(self, category, _filter):

        if category not in self._dict:
            self._dict[category] = {}

        # the filter can be registered as a tuple: ('filtername',filterfunction)
        # or just as a function in which case the function name is used as an
        # alias
        if isType(_filter, tuple) and len(_filter) >= 2:
            filtername = _filter[0]
            filterfunc = _filter[1]
        else:
            try:
                filtername = getName(_filter)
                filterfunc = _filter
            except AttributeError as e:
                raise ValueError(
                    'FilterManager: Invalid component filter %s.' % _filter)

        if filtername in self._dict[category]:
            raise ValueError('FilterManager: %s component filter already exists for %s category ' % (filtername, category))

        if category not in config.options:
            config.addOption(category, "", "")
        config.overrideDefaultValue(category, filtername)
        self._dict[category][filtername] = filterfunc

    def setDefault(self, _filter):
        if self.default:
            raise ValueError('FilterManager: default filter already exists')

        self.default = _filter

    def __getitem__(self, category):
        try:
            filters = self._dict[category]
        except KeyError:
            # no filters registered for this category
            if self.default:
                return self.default
            return void_filter

        try:
            filtername = config[category]
            return filters[filtername]
        except ConfigError:
            # if we have only one filter registered for this category we use it
            if len(filters) == 1:
                return filters.values()[0]
            else:  # ambiguity
                raise ValueError('FilterManager: Multiple filters detected for %s category: %s, '
                                 'but no one has be set as default in [GPIComponentFilters] section of the configuration file'
                                 % (category, str(filters.keys())))
        except KeyError:
            # wrong filter name in configuration for this category
            raise ValueError('FilterManager: %s filter is not registered for %s category.'
                             'Check your [GPIComponentFilters] section of the configuration file'
                             % (filtername, category))


# all filters register here...
allComponentFilters = _ComponentFilterManager()



from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

class JediRequirements(GangaObject):
    '''Helper class to group Jedi requirements.
    '''

    _schema = Schema(Version(1,1), {
        'long'          : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Send job to a long queue'),
        'cloud'         : SimpleItem(defvalue='US',protected=0,copyable=1,doc='cloud where jobs are submitted (default:US)'),
        'anyCloud'      : SimpleItem(defvalue=True,protected=0,copyable=1,doc='Set to true to allow jobs to run in all clouds. If False, jobs are limited to run in "requirements.cloud"'),
        'memory'        : SimpleItem(defvalue=-1,protected=0,copyable=1,doc='Required memory size'),
        'cputime'       : SimpleItem(defvalue=-1,protected=0,copyable=1,doc='Required CPU count in seconds. Mainly to extend time limit for looping detection'),
        'corCheck'      : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Enable a checker to skip corrupted files'),        
        'notSkipMissing': SimpleItem(defvalue=False,protected=0,copyable=1,doc='If input files are not read from SE, they will be skipped by default. This option disables the functionality'),
        'excluded_sites': SimpleItem(defvalue = [],typelist=['str'],sequence=1,protected=0,copyable=1,doc='Panda sites to be excluded while brokering.'),
        'excluded_clouds': SimpleItem(defvalue = [],typelist=['str'],sequence=1,protected=0,copyable=1,doc='Clouds to be excluded while brokering.'),
        'express'       : SimpleItem(defvalue = False,protected=0,copyable=1,doc='Send the job using express quota to have higher priority. The number of express subjobs in the queue and the total execution time used by express subjobs are limited (a few subjobs and several hours per day, respectively). This option is intended to be used for quick tests before bulk submission. Note that buildXYZ is not included in quota calculation. If this option is used when quota has already exceeded, the panda server will ignore the option so that subjobs have normal priorities. Also, if you submit 1 buildXYZ and N runXYZ subjobs when you only have quota of M (M < N),  only the first M runXYZ subjobs will have higher priorities'),
        'enableJEM'     : SimpleItem(defvalue = False,protected=0,copyable=1,doc='Enable the Job Execution Monitor.'),
        'configJEM'     : SimpleItem(defvalue = '',protected=0,copyable=1,doc='Config string for the Job Execution Monitor.'),
        'enableMerge'   : SimpleItem(defvalue = False, protected=0, copyable=1, doc='Enable the output merging jobs.'),
        'configMerge'   : SimpleItem(defvalue = {'type':'','exec':''}, protected=0, copyable=1, doc='Config parameters for output merging jobs.'),
        'usecommainputtxt' : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Boolean if input.txt contains the input files as comma separated list or separeted by a line break'),
        'rootver' : SimpleItem(defvalue = '',protected=0,copyable=1,doc='Specify a different root version for non-Athena jobs.'),
        'overwriteQueuedata'     : SimpleItem(defvalue = False,protected=0,copyable=1,doc='Expert option: overwriteQueuedata.'),
        'overwriteQueuedataConfig'     : SimpleItem(defvalue = '',protected=0,copyable=1,doc='Expert option: overwriteQueuedataConfig.'),
        'disableAutoRetry' : SimpleItem(defvalue=False,protected=0,copyable=1,doc='disable automatic job retry on the server side'),
        'noEmail' : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Suppress email notification'),
        'skipScout' : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Skip scout jobs'),
        'split'         : SimpleItem(defvalue=0,typelist=['type(None)','int'],copyable=1,doc='Number of sub-jobs to which a job is split.'),
        'nFilesPerJob'  : SimpleItem(defvalue=0,typelist=['type(None)','int'],copyable=1,doc='Number of files on which each sub-job runs.'),
        'nEventsPerFile'  : SimpleItem(defvalue=0,typelist=['type(None)','int'],copyable=1,doc='Number of events per file to run over'),
        'nGBPerJob'  : SimpleItem(defvalue=0,typelist=['type(None)','int'],copyable=1,doc='Instantiate one sub job per NGBPERJOB GB of input files. nGBPerJob=MAX sets the size to the default maximum value'),
        'maxNFilesPerJob'  : SimpleItem(defvalue=200,typelist=['type(None)','int'],copyable=1,doc='The maximum number of files per job is 200 by default since too many input files result in a too long command-line argument on WN which crashes the job. This option relax the limit.'),
    })

    _category = 'JediRequirements'
    _name = 'JediRequirements'
    _defcloud = ''
    
    def __init__(self):
        super(JediRequirements,self).__init__()
        from pandatools import PsubUtils
        import sys
        sys.stdout = open('/dev/null','w')
        sys.stderr = open('/dev/null','w')

        # cache the call to getCloudUsingFQAN as otherwise it can be slow
        if JediRequirements._defcloud == '':
            JediRequirements._defcloud = PsubUtils.getCloudUsingFQAN(None,False)
        self.cloud = self._defcloud

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        if not self.cloud:
            import random
            from pandatools import Client
            clouds = Client.PandaClouds.keys()
            clouds.remove('OSG')
            self.cloud = random.choice(clouds)



#! /usr/bin/env python
#
# ganga-readlfc.py
#
#
import os, re, commands, sys, getopt, signal

EC_Configuration = 410100

try:
    # ignore Python C API version mismatch
    sys.stderr = open("/dev/null", "w")
    # import
    import lfc
except:
    print 'ERROR: Could not import lfc' 
    sys.exit(EC_Configuration)

# repair stderr
sys.stderr = sys.__stderr__

try:
    sys.stderr = open("/dev/null", "w")
    import dq2.info.TiersOfATLAS as TiersOfATLAS
    from dq2.clientapi.DQ2 import DQ2
    from dq2.common.dao.DQDaoException import DQDaoException
    from dq2.info.TiersOfATLAS import _refreshToACache, ToACache
except:
    print 'ERROR: Could not import DQ2' 
    sys.exit(EC_Configuration)

# repair stderr
sys.stderr = sys.__stderr__

import random

_refreshToACache()
verbose = False

def handler(signum, frame):
    print "lfc.lfc_getreplicas timeout!"
    
##############################
def setLfcHost(lfcstring):
    # Set the LFC info
    # log.info('LFC endpoint is %s'%lfcstring)
    # Must be of the form  lfc://lfc0448.gridpp.rl.ac.uk:/grid/atlas
    lfc_re = re.compile('^lfc://([\w\d\-\.]+):([\w\-/]+$)')
    lfcre = lfc_re.match(lfcstring)
    # Set just the environ. It`s used be lfc api
    if lfcre:
        lfctup = {}
        lfctup['host'] = lfcre.group(1)
        realtime = commands.getoutput('date')
        #print realtime
        #print 'Setting env LFC_HOST to %s'%lfcre.group(1)
        os.environ['LFC_HOST'] = lfcre.group(1)
        # Root path, usually '/grid/atlas'
        lfctup['root'] = lfcre.group(2)
        # log.info('Setting env LFC_HOME to %s'%lfcre.group(2))
        os.environ['LFC_HOME'] = lfcre.group(2)
    else:
        print 'Cannot parse LFC string: %s'%lfcstring
        return 'WRAPLCG_WNCHECK_LFCSTRING'
    # Force lcg-utils to use LFC
    os.environ['LCG_CATALOG_TYPE']='lfc'
    return 'OK'

##############################
def getreplicas_bulk(guids, allLFCs):
    """
    allLFCs=[]
    for cl in TiersOfATLAS.ToACache.dbcloud:
        id=TiersOfATLAS.ToACache.dbcloud[cl]
        l = TiersOfATLAS.getLocalCatalog(id)
        if l:
            allLFCs += [l]
    # shuffle these
    random.shuffle(allLFCs)
    # Add the old central LFC last - only use if no-other replica found
    centralLFC = TiersOfATLAS.getLocalCatalog('OLDLCG')
    allLFCs += [ centralLFC ]
    #print allLFCs
    """
    
    guidSizes    = {}
    guidReplicas = {}
    guidmd5sum = {}
    for lfcstring in allLFCs:
        if setLfcHost(lfcstring) == 'OK':  
            print lfcstring
            sys.stdout.flush()
            signal.signal(signal.SIGQUIT, handler)
            signal.alarm(30)
            try:
                (res, rep_entries) = lfc.lfc_getreplicas(guids, '')
            except AttributeError:
                print 'ERROR LCG UI does not support LFC bulk reading - please upgrade !'
                sys.exit(EC_Configuration)
                
            signal.alarm(0)
                   
            if res != 0:  
                print 'lfc_getreplicas :  Error ' + str(res) + "\n"
            else:
                done = False  
                for rep in rep_entries:
                    if rep.errcode == 0 and rep.sfn!='':
                        if rep.sfn.startswith("srm"):

                            # Add 8443 port
                            surl = rep.sfn.split("//")
                            surl[1]= rep.sfn.split("//")[1].split("/")[0]+":8443/"+'/'.join(rep.sfn.split("//")[1].split("/")[1:])
                            surl_8443 = '//'.join(surl)
                            # Use original sfn
                            surl_8443 = rep.sfn
                        else:
                            surl = rep.sfn.split("//")
                            surl[0]= 'gsiftp:'
                            surl[1]= rep.sfn.split("//")[1].split("/")[0]+":2811/"+'/'.join(rep.sfn.split("//")[1].split("/")[1:])
                            surl_8443 = '//'.join(surl)
          
                        surl_8443 = re.sub(':*\d*/srm/managerv2\?SFN=','', surl_8443 )
                        if rep.guid in guidReplicas.keys():
                            guidReplicas[rep.guid].append(surl_8443) 
                        else:
                            guidSizes[rep.guid] = rep.filesize
                            #guidReplicas[rep.guid] = [rep.sfn] 
                            guidReplicas[rep.guid] = [surl_8443]
                            guidmd5sum[rep.guid] = rep.csumvalue
  
    return 'OK',guidReplicas, guidSizes,guidmd5sum

##############################
def getinputreplicas(lfn_guid, allLFCs):

    exitcode, guidReplicas,guidSizes,guidmd5sum = getreplicas_bulk(lfn_guid.values(), allLFCs)

    for lfn,guid in lfn_guid.items():
        if guid not in guidReplicas.keys():
            #print 'Fail to get replicas for inputfile '+guid
            #print 'Try single query. '
            return 'Not Found',  guidReplicas,guidSizes,guidmd5sum
    
    retcode = pre_stagein(guidReplicas)
    if retcode !='OK':
        return 'Not Pre-stage', guidReplicas,guidSizes,guidmd5sum        
    else:
        return 'OK', guidReplicas,guidSizes,guidmd5sum
    
##############################
def pre_stagein(guidReplicas):

    site = commands.getoutput('/bin/hostname')
    if site.find('cern.ch'):
        for guid in guidReplicas.keys():
            for surl in guidReplicas[guid]:
                if surl.split('//')[1].startswith('srm.cern.ch'):
                    #file need to be pre-stagein
                    castor_path = '/'+'/'.join(surl.split('//')[1].split("/")[1:])
                    cmd = 'stager_get -M '+castor_path
                    (exitcode, output) = commands.getstatusoutput(cmd)
                    if exitcode == 0: # if fail to stager_get
                        cmd = 'stager_qry -M '+castor_path
                        (exitcode, output) = commands.getstatusoutput(cmd)
                        if exitcode == 0: # if fail to stager_qry
                            print castor_path+' has been pre-staged.'        
                        else:
                            return 'Not OK' 
                    else:
                        return 'Not OK'                  
                    
                    break
        return 'OK'
    else:
        "For submit node out of cern, pre-stagein is not needed." 

#####################################
def _usage():
    print """ganga-readlfc.py datasetname complete=0/1"""

#####################################
if __name__ == '__main__':

    dataset = ''
    complete = -1
    removefromlfclist = [ ]
    list = False
    list_guids = False

    try:
        opts, args = getopt.getopt(sys.argv[1:],"hvr:lg", ["help","verbose","remove","list","list_guids"])
    except:
        _usage()
        print "ERROR : Invalid options"
        sys.exit(EC_Configuration)    

    # set options
    for o, a in opts:
        if o in ("-h","--help"):
            _usage()
            sys.exit(EC_Configuration)
        if o in ("-v","--verbose"):
            verbose = True
        if o in ("-l","--list"):
            list = True
        if o in ("-g","--list_guids"):
            list = True
            list_guids = True
        if o in ("-r","--remove"):
            removefromlfclist = a.split(",")

    if args and len(args)>1:
        dataset = args[0]
        complete = int(args[1])
    else:
        print 'ERROR no dataset name given !'
        sys.exit(EC_Configuration)

    dq2=DQ2()
    content = dq2.listDatasets(dataset)
    if content=={}:
        print 'ERROR Dataset %s is not defined in DQ2 database !' %dataset
        sys.exit(0)
                                                                   

    contents = dq2.listFilesInDataset(dataset)
    if contents:
        contents = contents[0]
    else:
        contents = {}

    if complete==1:
        locations = dq2.listDatasetReplicas(dataset, complete=1)
    elif complete==0:
        locations = dq2.listDatasetReplicas(dataset, complete=0)
    else:
        locations = dq2.listDatasetReplicas(dataset)
    datasetinfo = dq2.listDatasets(dataset)
    datasetvuid = datasetinfo[dataset]['vuids'][0]

    if locations:
        locations = locations[datasetvuid]
    else:
        print 'ERROR no location'
        sys.exit(0)

    if complete == -1: 
        locrange = xrange(0,2)
    elif complete == 0: 
        locrange = xrange(0,1)
    elif complete == 1:
        locrange = xrange(1,2)

    locnum = 0
    for i in locrange:
        locnum += len(locations[i])
    if locnum == 0:
        print 'ERROR no location'
        sys.exit(0)
            
    lfn_guid = {}
    for guid, info in contents.iteritems():
        lfn_guid[info['lfn']] = guid

    allLFCs = []


    for i in locrange:
        for location in locations[i]:
            l = TiersOfATLAS.getLocalCatalog(location)
            if l and l not in allLFCs and l.startswith('lfc') and l not in removefromlfclist:
                allLFCs.append(l)
            

    status, guidReplicas, guidSizes, guidmd5sum = getinputreplicas(lfn_guid, allLFCs)

    print guidReplicas
    locations_srm = {}
    for i in locrange:
        for location in locations[i]:
            try:
                if 'srm' in TiersOfATLAS.ToACache.sites[location]:
                    tempsrm = TiersOfATLAS.ToACache.sites[location]['srm']
                    tempsrm = re.sub('token:*\w*:','',tempsrm)
                    tempsrm = re.sub(':*\d*/srm/managerv2\?SFN=','', tempsrm)
                    print tempsrm
                    locations_srm[location]=tempsrm
            except KeyError:
                pass
            
    locations_num = {}
    location_list = {}
    guid_list = {}
    for i in locrange:
        for location in locations[i]:
            locations_num[location]=0
            location_list[location] = []

    for guid, surllist in guidReplicas.iteritems():
        for location, location_srm in locations_srm.iteritems():
            for surl in surllist:
                if surl.startswith(location_srm):
                    locations_num[location]=locations_num[location]+1
                    break
                elif location=='DESY-HH' and surl.startswith("srm://srm-dcache.desy.de"):
                    locations_num['DESY-HH']=locations_num['DESY-HH']+1
                    break
                
    for guid, surllist in guidReplicas.iteritems():
        guid_list[guid] = []
        for location, location_srm in locations_srm.iteritems():
            for surl in surllist:
                if surl.startswith(location_srm):
                    item = [ guid, surl ]
                    newitem = location_list[location] + item 
                    location_list[location] = newitem
                    guid_list[guid].append( location )

    if not list:
        for location, num in locations_num.iteritems():
            print '#%s:%s' %(location, num)
            sys.stdout.flush()
    else:
        if not list_guids:
            for location, file in location_list.iteritems():
                line = '#'+location
                for ifile in file: 
                    line = line + ',%s' %ifile
                print line
                sys.stdout.flush()
        else:
            for guid, locations in guid_list.iteritems():
                line = '#'+guid
                for location in locations: 
                    line = line + ',%s' %location
                print line
                sys.stdout.flush()



from Ganga.Core.exceptions import GangaException, BackendError
#from GangaDirac.BOOT       import dirac_ganga_server
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from Ganga.Utility.logging import getLogger
logger = getLogger()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def result_ok(result):
    '''
    Check if result of DIRAC API command is OK.
    '''
    if result is None:
        return False
    elif type(result) is not dict:
        return False
    else:
        output = result.get('OK', False)
        return output


def get_result(command,
               logger_message=None,
               exception_message=None,
               eval_includes=None,
               retry_limit=5):

    retries = 0
    while retries < retry_limit:

        try:
            result = execute(command, eval_includes=eval_includes)

            if not result_ok(result):
                if logger_message is not None:
                    logger.warning('%s: %s' % (logger_message, str(result)))
                if exception_message is not None:
                    logger.warning("Failed to run: %s" % str(command))
                    logger.warning("includes:\n%s" % str(eval_includes))
                    logger.warning("Result: '%s'" % str(result))
                    raise GangaException(exception_message)
                raise GangaException("Failed to return result of '%s': %s" % (command, result))
            return result
        except Exception as x:
            import time
            logger.debug("Sleeping for 5 additional seconds to reduce possible overloading")
            time.sleep(5.)
            if retries == retry_limit - 1:
                raise x
            retries = retries + 1
            logger.error("An Error Occured: %s" % str(x))
            logger.error("Retrying: %s / %s " % (str(retries + 1), str(retry_limit)))


def get_job_ident(dirac_script_lines):
    '''parse the dirac script for the label given to the job object'''
    target_line = [
        line for line in dirac_script_lines if line.find('Job()') >= 0]
    if len(target_line) != 1:
        raise BackendError(
            'Dirac', 'Could not determine the identifier of the Dirac Job object in API script')

    return target_line[0].split('=', 1)[0].strip()


def get_parametric_datasets(dirac_script_lines):
    '''parse the dirac script and retrieve the parametric inputdataset'''
    method_str = '.setParametricInputData('

    def parametric_input_filter(API_line):
        return API_line.find(method_str) >= 0
        # return API_line.find('.setParametricInputData(') >= 0

    parametric_line = filter(parametric_input_filter, dirac_script_lines)
    if len(parametric_line) is 0:
        raise BackendError(
            'Dirac', 'No "setParametricInputData()" lines in dirac API')
    if len(parametric_line) > 1:
        raise BackendError(
            'Dirac', 'Multiple "setParametricInputData()" lines in dirac API')

    end_method_marker = parametric_line[0].find(method_str) + len(method_str)
    dataset_str = parametric_line[0][end_method_marker:-1]
    return eval(dataset_str)


# Note could combine selection_pred with file_type
# using types.typetype or types.functiontype
def outputfiles_iterator(job, file_type, selection_pred=None,
                         include_subfiles=True):
    def combined_pred(f):
        if selection_pred is not None:
            return issubclass(f.__class__, file_type) and selection_pred(f)
        return issubclass(f.__class__, file_type)

    import itertools
    for f in itertools.chain(job.outputfiles, job.non_copyable_outputfiles):
        if include_subfiles and hasattr(f, 'subfiles') and f.subfiles:
            for sf in itertools.ifilter(combined_pred, f.subfiles):
                yield sf
        else:
            if combined_pred(f):
                yield f


def outputfiles_foreach(job, file_type, func, fargs=(), fkwargs=None,
                        selection_pred=None, include_subfiles=True):
    if fkwargs is None:
        fkwargs = {}
    if fargs is None:
        fargs =  ()
    output = []
    for f in outputfiles_iterator(job, file_type, selection_pred, include_subfiles):
        output.append(func(f, *fargs, **fkwargs))
    return output

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def ifilter_chain(selection_pred, *iterables):
    import itertools
    for item in itertools.ifilter(selection_pred,
                                  itertools.chain(*iterables)):
        yield item


def for_each(func, *iterables, **kwargs):
    result = []
    for item in ifilter_chain(kwargs.get('selection_pred', None),
                              *iterables):
        result.append(func(item,
                           *kwargs.get('fargs', ()),
                           **kwargs.get('fkwargs', {})))
    return result



"""Core finisher IAction implementation.

The CoreFinisher class provides a generic finisher implementation.
 
"""

from GangaRobot.Lib.Base.BaseFinisher import BaseFinisher


class CoreFinisher(BaseFinisher):
    
    """Core finisher IAction implementation.
    
    A finisher which evaluates if jobs are finished in a generic way. See
    handleisfinished() for details.
    
    This finisher can be reused by adding it to the chain of finishers
    initialised in the constructor of any implementation of BaseFinisher.
    e.g.
    def __init__(self):
        self.chain = [CoreSubmitter(), self]
    
    """

    def handleisfinished(self, job):
        """Evaluate if job is finished based on job status.
        
        Keyword arguments:
        job -- A Ganga Job associated with the run.
        
        Return True if job status is 'completed', 'killed', 'failed',
        'unknown', or 'new' (i.e. if reverted to 'new' after submission failed).
        
        Return False otherwise.
        
        """
        return job.status in ['completed', 'killed', 'failed', 'unknown', 'new']
    


import os

from Ganga.Runtime.GPIexport import exportToGPI
from Ganga.GPIDev.Base.Proxy import GPIProxyClassFactory

#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
from GangaLHCb.Lib.Applications import AppsBaseUtils
from Ganga.Utility.logging import getLogger

from GaudiPython import GaudiPython

from Bender import Bender

# Add any additional Packages required by the user in the .gangarc file
from Ganga.Utility.Config import getConfig

logger = getLogger()
logger.debug("User Added Apps")

config = getConfig('LHCb')
user_added = config['UserAddedApplications']
user_apps = user_added.split(':')
if user_apps == user_added and len(user_added) > 0:
    AppsBaseUtils.addNewLHCbapp(user_apps)
for app in user_apps:
    if len(app) > 0:
        AppsBaseUtils.addNewLHCbapp(app)

logger.debug("Constructing AppsBase Apps")
f = open(os.path.join(os.path.dirname(__file__), 'AppsBase.py'), 'r')
cls = f.read()
f.close()
all_apps = ''
for app in AppsBaseUtils.available_apps():
    if app in dir():
        continue
    app = str(app)
    this_app = cls.replace('AppName', app)
    all_apps = all_apps + str('\n\n') + this_app
    #exec(this_exec, all_global, all_local)
    logger.debug("Adding %s" % str(app))

logger.debug("Adding apps")
modules= compile(all_apps, '<string>', 'exec')
exec modules

logger.debug("Fin")





from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.Utility.Config import getConfig
from Ganga.Utility.files import expandfilename
from Ganga.GPIDev.Lib.Job import Job
import pickle
import os
import copy
import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

class GaussSplitter(ISplitter):

    """Create a set of Gauss jobs based on the total number of jobs and the
    number of events per subjob.

    This Splitter will create a set of Gauss jobs using two parameters:
    'eventsPerJob' and 'numberOfJobs'. Each job uses a different random seed
    using the Gaudi options file statement 'GaussGen.FirstEventNumber' and will
    produce the amount of events sepcified in 'eventsPerJob'. The total number
    of generated events therefore will be 'eventsPerJob*numberOfJob'.
    """
    _name = "GaussSplitter"
    _schema = Schema(Version(1, 0), {
        'eventsPerJob': SimpleItem(defvalue=5, doc='Number of generated events per job'),
        'numberOfJobs': SimpleItem(defvalue=2, doc="No. of jobs to create"),
        'firstEventNumber': SimpleItem(defvalue=0, doc="First event number for first subjob")
    })

    def _create_subjob(self, job, inputdata):
        j = Job()
        j.copyFrom(job)
        j.splitter = None
        j.merger = None
        j.inputsandbox = []  # master added automatically
        j.inputfiles = []
        j.inputdata = inputdata

        return j

    def split(self, job):
        subjobs = []

        inputdata = job.inputdata
        if not job.inputdata:
            share_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),
                                      'shared',
                                      getConfig('Configuration')['user'],
                                      job.application.is_prepared.name,
                                      'inputdata',
                                      'options_data.pkl')

            if os.path.exists(share_path):
                f = open(share_path, 'r+b')
                inputdata = pickle.load(f)
                f.close()

        for i in range(self.numberOfJobs):
            j = self._create_subjob(job, inputdata)
            first = self.firstEventNumber + i * self.eventsPerJob + 1
            opts = 'from Gaudi.Configuration import * \n'
            opts += 'from Configurables import GenInit \n'
            opts += 'ApplicationMgr().EvtMax = %d\n' % self.eventsPerJob
            #opts += 'from Configurables import LHCbApp\n'
            #opts += 'LHCbApp().EvtMax = %d\n' % self.eventsPerJob
            opts += 'GenInit("GaussGen").FirstEventNumber = %d\n' % first
            spillOver = ["GaussGenPrev", "GaussGenPrevPrev", "GaussGenNext"]
            for s in spillOver:
                opts += 'GenInit("%s").FirstEventNumber = %d\n' % (s, first)
            #j.application.extra.input_buffers['data.py'] += opts
            j._splitter_data = opts
            # j.inputsandbox.append(File(FileBuffer(path,opts).create().name))
            logger.debug("Creating job %d w/ FirstEventNumber = %d" % (i, first))
            subjobs.append(j)

        return subjobs




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
"""The LHCb Bookkeeping interface to Ganga."""

import os
import sys
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
import Ganga.Utility.logging

logger = Ganga.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

schema = {'status': SimpleItem(defvalue=None, protected=1, copyable=0,
                               typelist=['str', 'type(None)'],
                               doc='Status of the bookkeeping system')}


class Bookkeeping(GangaObject):
    _schema = Schema(Version(1, 0), schema)
    _exportmethods = ['browse']
    _category = 'datasets'
    _name = 'Bookkeeping'

    def __init__(self):
        super(Bookkeeping, self).__init__()
        pass

    def _createTmpFile(self):
        import tempfile
        temp_fd, temp_filename = tempfile.mkstemp(text=True, suffix='.txt')
        os.write(temp_fd, '')
        os.close(temp_fd)
        return temp_filename

    def browse(self, gui=True):
        f = self._createTmpFile()
        if gui:
            cmd = 'bookkeepingGUI("%s")' % f
            execute(cmd)
            l = self._fileToList(f)
            ds = LHCbDataset()
            ds.__construct__([l])
            return ds

    def _fileToList(self, file):
        f = open(file)
        l = f.read().splitlines()
        # prefix files with 'LFN:' to make ganga think they are lfns
        for i in range(len(l)):
            if not l[i].upper().startswith('LFN'):
                l[i] = 'LFN:' + os.path.normpath(l[i])
        f.close()
        return l

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
from Ganga.GPIDev.Lib.Job import Job
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.Utility.files import expandfilename
import Ganga.Utility.logging
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
import Ganga.Utility.Config
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.GPIDev.Lib.File import FileBuffer, File
from Ganga.Core import TypeMismatchError
from Ganga.Utility.util import unique
import shutil
from RunTimeHandlerUtils import sharedir_handler
from Ganga.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import master_sandbox_prepare, sandbox_prepare
logger = Ganga.Utility.logging.getLogger()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class GaudiRunTimeHandler(IRuntimeHandler):

    """This is the application runtime handler class for Gaudi applications 
    using the local, interactive and LSF backends."""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig, ['inputsandbox'])
        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):

        inputsandbox, outputsandbox = sandbox_prepare(
            app, appsubconfig, appmasterconfig, jobmasterconfig)

        run_script = self.__create_run_script(app,
                                              appsubconfig,
                                              appmasterconfig,
                                              jobmasterconfig,
                                              inputsandbox,
                                              outputsandbox)
        return StandardJobConfig(FileBuffer('gaudi-script.py', run_script, executable=1),
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def __create_run_script(self,
                            app,
                            appsubconfig,
                            appmasterconfig,
                            jobmasterconfig,
                            inputsandbox,
                            outputsandbox):
        job = app.getJobObject()

        opts = 'options.pkl'

        script = "#!/usr/bin/env python\n\nimport os,sys\n\n"
        script += 'opts = \'%s\'\n' % opts
        script += 'app = \'%s\'\n' % app.appname
        script += 'app_upper = \'%s\'\n' % app.appname.upper()
        script += 'version = \'%s\'\n' % app.version
        script += 'package = \'%s\'\n' % app.package
        script += 'platform = \'%s\'\n' % app.platform

        if opts:
            script += """# check that options file exists
if not os.path.exists(opts):
    opts = 'notavailable'
    os.environ['JOBOPTPATH'] = opts
else:
    os.environ['JOBOPTPATH'] = '%s/%s/%s_%s/%s/%s/%s/options/job.opts' \
                               % (os.environ[app + '_release_area'],app_upper,
                                  app_upper,version,package,app,version)
    import sys
    sys.stdout.write('Using the master optionsfile: '+ str(opts))
    sys.stdout.flush()
    
"""

        script += """# check that SetupProject.sh script exists, then execute it

# add lib subdir in case user supplied shared libs where copied to pwd/lib
os.environ['LD_LIBRARY_PATH'] = '.:%s/lib:%s\' %(os.getcwd(),
                                                 os.environ['LD_LIBRARY_PATH'])
                                                 
#run
sys.stdout.flush()
os.environ['PYTHONPATH'] = '%s/InstallArea/python:%s' % \\
                            (os.getcwd(), os.environ['PYTHONPATH'])
os.environ['PYTHONPATH'] = '%s/InstallArea/%s/python:%s' % \\
                            (os.getcwd(), platform,os.environ['PYTHONPATH'])

cmdline = \"\"\"gaudirun.py '
"""

        for arg in app.args:
            script += arg + ' '
        script += '%s data.py\"\"\" % opts \n'

        script += """
# run command
os.system(cmdline)
"""
        return script, inputsandbox, outputsandbox


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



from ATLASDataset import *
from ATLASCastorDataset import *
from DQ2Dataset import *
from StagerJobSplitter import *
from StagerDataset import *
from EventPicking import *
from DQ2SandboxCache import *



from __future__ import print_function, absolute_import

##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: __init__.py,v 1.7 2009-06-25 13:30:00 moscicki Exp $
##########################################################################

#
# Thin wrapper for logging subsystem:
#  - makes standard python 2.3 logging available in python 2.2
#  - integrates loggers with Ganga configuration subsystem in a most covinient way
#  - extra reporting utilities:
#     log_user_exception
#
# Design principles:
#  - getLogger() returns the standard logger object for the current execution context
#     i.e. with the name which indicates the current package or module name
#  - all loggers are automatically configured according to this modules config dictionary (see below)
#  - special functions:
#       - log_user_exception() allows to format nicely exception messages

import cStringIO
import logging
import logging.handlers
import os.path
import sys
import threading
import traceback

# logger configuration
# settings for new loggers may be added here at will (for example read
# from the config file)

import Ganga.Utility.ColourText as ColourText

from Ganga.Utility.Config import getConfig
config = getConfig("Logging")

# initialize the root logger for the logger created directly in python
# executable scripts which have no name starting by "Ganga."
# By default everything goes to stdout

_hasInit = False

if not _hasInit:
    logging.basicConfig(stream=sys.stdout)
    _hasIinit = True

_formats = {
    'DEBUG': '%(asctime)s %(threadName)s %(module)-20s::%(funcName)-20s:%(lineno)d %(levelname)-8s: %(message)s',
    'VERBOSE': '%(asctime)s %(module)-25s::%(funcName)-10s: %(levelname)-8s %(message)s',
    'NORMAL': '%(levelname)-8s %(message)s',
    'TERSE': '%(levelname)-8s %(message)s',
}

requires_shutdown = False

private_logger = None  # private logger of this module

# main logger corresponds to the root of the hierarchy
main_logger = logging.getLogger()

# this is the handler used to print on screen directly
direct_screen_handler = main_logger.handlers[0]  # get default StreamHandler

# if defined this is the handler used for caching background messages at
# interactive prompt
cached_screen_handler = None

# this is the handler currenty in use by main_logger (either direct_screen_handler or cached_screen_handler)
# or it may be overriden by bootstrap() to be arbitrary handler
default_handler = direct_screen_handler

# This is the default formatter key from _formats which is to be used in constructing new Handlers
# None by default otherwise it interferes with user configs
default_formatter = None

# if defined this is ADDITIONAL handler that is used for the logfile
file_handler = None


# FIXME: this should be probably an option in the config
# if defined, global level (string) overrides anything which is in config
_global_level = None

# all loggers which are used by all modules
_allLoggers = {}


# use this function to get new loggers into your packages
# if you do not provide the name then logger will detect your package name
# if you specify the modulename as a string then it will be appended to the package name
# if you specify the modulename==1 then your module name will be guessed and appended to the package name
# the guessing algorithm may be modified by passing the frame object (to emulate a different physical location of the logger)
# this is only useful for special usage such as IBackend base class
def getLogger(name=None, modulename=None):
    """ get a logger depending on if the modulename is 1/0 and if the name of the logger is provided, see _getLogger for more details"""
    return _getLogger(name, modulename)


class ColourFormatter(logging.Formatter, object):
    """ Adds colour to our formatting """

    def __init__(self, *args, **kwds):
        logging.Formatter.__init__(self, *args, **kwds)
        fg = ColourText.Foreground()
        fx = ColourText.Effects()
        self.colours = {logging.INFO: fx.normal,
                        logging.WARNING: fg.orange,
                        logging.ERROR: fg.red,
                        logging.CRITICAL: fg.red,
                        logging.DEBUG: fx.normal}
        self.markup = ColourText.ANSIMarkup()

    def format(self, record):
        """ Actually format this message(record) and return the formatted version"""
        try:
            s = super(ColourFormatter, self).format(record)
        except TypeError:
            print("%s" % str(record))
            return None
        if record.levelno in self.colours.keys():
            code = self.colours[record.levelno]
            return self.markup(s, code)
        else:
            return s

    def setColour(self, yes):
        """ Boolean set for if teh colours should be used"""
        if yes:
            self.markup = ColourText.ANSIMarkup()
        else:
            self.markup = ColourText.NoMarkup()


def _set_formatter(handler, this_format=None):
    """ Set the formatter for this handler. 1 formatter per handler.
    Precdence in deciding which format to use
    _customFormat in the Config changes all formats all the time
    this_format here overloads the default in the config (useful for --debug
    otherwise the format set by default in the config is used"""
    if config['_customFormat'] != "":
        for k in _formats.keys():
            _formats[k] = config['_customFormat']

    if this_format is not None and this_format in _formats:
        formatter = ColourFormatter(_formats[this_format])
    else:
        formatter = ColourFormatter(_formats[config['_format']])
    formatter.setColour(config['_colour'])
    handler.setFormatter(formatter)


def _make_file_handler(logfile, logfile_size):
    """ Make a new file handler and add it to the main_logger.
    Makes use of logfile and logfile_size to know where to put the log and how big to ket it get"""
    logfile = os.path.expanduser(logfile)
    global file_handler
    if logfile:
        try:
            new_file_handler = logging.handlers.RotatingFileHandler(
                logfile, maxBytes=logfile_size, backupCount=1)
        except IOError as x:
            private_logger.error('Cannot open the log file: %s', str(x))
            return
        # remove old handler if exists
        if file_handler:
            main_logger.removeHandler(file_handler)
            file_handler.flush()
            file_handler.close()
            # this is required to properly remove the file handler from the logging system
            # otherwise I/O Error at shutdown
            try:
                # WARNING: this relies on the implementation details of the
                # logging module
                del logging._handlers[file_handler]
            except KeyError:
                # don't complain if the handler was correctly unregistered by
                # the logging system
                pass

        new_file_handler.setFormatter(logging.Formatter(_formats['DEBUG']))
        main_logger.addHandler(new_file_handler)
        file_handler = new_file_handler


# set the loglevel for a logger to a given string value (example: "DEBUG")
def _set_log_level(logger, value):
    """ Sets the log level for this logger. When in info, only display INFO/WARNING/... etc the value is the level to set, must be a standard Python logging level """

    global _global_level

    if _global_level is not None:
        value = _global_level

    # convert a string "DEBUG" into enum object logging.DEBUG
    def _string2level(name):
        if hasattr(logging, name):
            return getattr(logging, name)

    try:
        logger.setLevel(_string2level(value))
        return value
    except AttributeError as err:
        logger.error('Attribute Error: %s', str(err))
        logger.warning('possible configuration error: invalid level value (%s), using default level', value)
        return None


# reflect all user changes immediately
def post_config_handler(opt, value):
    """ This is called after a config has been set upon startup in the Schema, make the changes here immediate, change 1 option (opt) to 1 value """

    if config is not None and '_customFormat' in config and config['_customFormat'] != "":
        for k in _formats.keys():
            _formats[k] = config['_customFormat']

    if config is not None:
        _format, colour = config['_format'], config['_colour']

    if opt in ['_format', '_customFormat']:
        badConfig = False
        if opt == "_customFormat":
            if config is not None:
                value = config['_format']
            else:
                value = None
        if _formats is not None and value in _formats:
            _format = _formats[value]
        else:
            if private_logger is not None:
                private_logger.error('illegal name of format string (%s), possible values: %s' % (str(value), _formats.keys()))
            return

    if opt == '_colour':
        colour = value

    if opt in ['_format', '_colour', '_customFormat']:
        fmt = ColourFormatter(_format)
        fmt.setColour(colour)
        direct_screen_handler.setFormatter(fmt)
        return

    if config is not None:
        logfile, logfile_size = config['_logfile'], config['_logfile_size']

    if opt in ['_logfile', '_logfile_size']:
        _make_file_handler(logfile, logfile_size)
        return

    # FIXME: has no effect at runtime, should raise a ConfigError
    if opt == '_interactive_cache':
        return

    # set the logger level
    if private_logger is not None:
        private_logger.info('setting loglevel: %s %s', opt, value)

    if _set_log_level is not None and getLogger is not None:
        _set_log_level(getLogger(opt), value)


config.attachUserHandler(None, post_config_handler)
config.attachSessionHandler(None, post_config_handler)

lookup_frame_names = {}


def _guess_module_logger_name(modulename, frame=None):
    """Gues the Module name from the current frame or from a given frame if specified. If module == 1 return full name, else trim module name"""
    # find the filename of the calling module
    if frame is None:
        # assuming 2 nested calls to the module boundary!
        frame = sys._getframe(3)
    else:
        print('using frame from the caller')

    global lookup_frame_names


    this__file__ = None
    if '__file__' in frame.f_globals.keys():
        this__file__ = frame.f_globals['__file__']
        if this__file__ in lookup_frame_names:
            del frame
            return lookup_frame_names[this_file]
        else:
            should_store = True
    else:
        should_store = False

    # accessing __file__ from globals() is much more reliable than
    # f_code.co_filename (name = os.path.normcase(frame.f_code.co_filename))
    if this__file__ is not None:
        name = os.path.realpath(os.path.abspath(this__file__))
    else:
        # no file associated with the frame (e.g. interactive prompt, exec
        # statement)
        name = '_program_'

    # print " _guess_module_logger_name",name
    del frame

    # if private_logger:
    #    private_logger.debug('searching for package matching calling module co_filename= %s',str(name))

    # sometimes the filename is an absolute path, try to find a relative module path from the PYTHONPATH
    # and remove the trailing path -> the result will be used as the logger
    # name

    from Ganga.Utility.files import remove_prefix

    name = remove_prefix(name, sys.path)

    def remove_tail(s, tail):
        idx = s.rfind(tail)
        if idx != -1:
            return s[:idx]
        return s

    # get rid of trailing .py  .pyc .pyo
    name = remove_tail(name, '.py')

    # replace slashes with dots
    name = name.replace(os.sep, '.')

    # return full module name
    if modulename == 1:
        return name

    # remove module name
    name = remove_tail(name, '.')

    if name == 'ganga':  # interactive IPython session
        name = "Ganga.GPI"

    # return package name
    if not modulename:
        return name

    return_name = name + '.' + modulename

    if should_store is True:
        lookup_frame_names[this__file__] = return_name

    # return custom module name
    return return_name

_MemHandler = logging.handlers.MemoryHandler


# Caching will not be done for messages which are generated by the main thread.
class FlushedMemoryHandler(_MemHandler):
    """ Flushed memory handler used for caching anything sent to the logging not on the MainThread"""

    def __init__(self, *args, **kwds):
        _MemHandler.__init__(self, *args, **kwds)

    def shouldFlush(self, record):
        """
        Right here is where we make the decision on what to buffer or not in the logger in the interactive mode.
        The only thread we don't want to buffer is the MainThread. All other threads are supporting threads and not of primary interest.
        The exception to this is when the MemHandler says we flush as we want to always flush then.
        """
        return (threading.currentThread().getName() == "MainThread") or \
                _MemHandler.shouldFlush(self, record)


def enableCaching():
    """
    Enable caching of log messages at interactive prompt. In the interactive IPython session, the messages from monitoring
    loop will be cached until the next prompt. In non-interactive sessions no caching is required.
    """

    if not config['_interactive_cache']:
        return

    private_logger.debug('CACHING ENABLED')
    global default_handler, cached_screen_handler
    main_logger.removeHandler(default_handler)
    cached_screen_handler = FlushedMemoryHandler(1000, target=direct_screen_handler)
    default_handler = cached_screen_handler
    if default_formatter is not None:
        _set_formatter(cached_screen_handler, default_formatter)
    main_logger.addHandler(default_handler)


def _getLogger(name=None, modulename=None):
    """ Get a logger, either by name or by modulename and return it"""

    if name is None:
        name = _guess_module_logger_name(modulename)

    if name.split('.')[0] != 'Ganga' and name != 'Ganga':
        name = 'Ganga.' + name

    if name in _allLoggers:
        return _allLoggers[name]
    else:

        logger = logging.getLogger(name)
        _allLoggers[name] = logger

        if name in config:
            thisConfig = config[name]
            _set_log_level(logger, thisConfig)

        return logger


# this bootstrap method should be called very early in the bootstrap phase of the entire system to enable
# the log level configuration to take effect...
# the optional handler overrides the default handler and is used for GUI environment
# NOTE: the additional message buffering handler is available only for the
# default handler

def bootstrap(internal=False, handler=None):
    """ This is called in the main bootstrap to initialize, the logfile, the main_logger and logging defaults
    This requires access to a fully loaded config system in order to correctly initialize all objects based upon .gangarc"""

    global private_logger, main_logger

    if internal is True and private_logger is not None:
        return

    private_logger = getLogger('Ganga.Utility.logging')
    #_set_log_level(private_logger, 'CRITICAL')
    #main_logger = _getLogger('Ganga',_roothandler=1,handler=handler)

    private_logger.debug('bootstrap')

    if internal:
        _make_file_handler(config['_logfile'], config['_logfile_size'])
    else:
        # override main_logger handler
        global default_handler
        if handler is not None and handler is not default_handler:
            main_logger.removeHandler(default_handler)
            main_logger.addHandler(handler)
            default_handler = handler

    if default_formatter is not None:
        _set_formatter(default_handler, default_formatter)
    else:
        _set_formatter(default_handler)
    # main_logger.propagate = 0 # do not propagate messages upwards...
    # main_logger.addHandler(default_handler)
    if file_handler:
        main_logger.addHandler(file_handler)

    opts = filter(lambda o: o.find('Ganga') == 0, config)
    for opt in opts:

        # should reconfigure Ganga and private logger according to the config
        # file contents

        msg = 'logging %s in %s mode' % (opt, config[opt])
        if internal:
            private_logger.debug(msg)
        else:
            private_logger.info(msg)

        _set_log_level(getLogger(opt), config[opt])

    if not internal:
        class NoErrorFilter(logging.Filter):
            """
            A filter which only allow messages which are WARNING or lower to be logged
            """
            def filter(self, record):
                return record.levelno <= 30

        # Make the default handler not print ERROR and CRITICAL
        direct_screen_handler.addFilter(NoErrorFilter())

        # Add a new handler for ERROR and CRITICAL which prints to stderr
        error_logger = logging.StreamHandler(sys.stderr)
        error_logger.setLevel(logging.ERROR)
        _set_formatter(error_logger)
        main_logger.addHandler(error_logger)

    global requires_shutdown
    requires_shutdown = True

    private_logger.debug('end of bootstrap')


def final_shutdown():
    """ Shutdown the logging system via a call here as we don't want to do this in the wrong place in the shutdown method """
    private_logger.debug('shutting down logsystem')
    logging.shutdown()


# do the initial bootstrap automatically at import -> this will bootstrap
# the main and private loggers
bootstrap(internal=True)

# force all loggers to use the same level


def force_global_level(level):
    """ Force the global logging level to be set to level. In the case of DEBUG change the formatting to reflect this """
    if level is not None:
        for l in _allLoggers:
            _set_log_level(_allLoggers[l], level)

        global _global_level
        _global_level = level

    ## May wish to do this for more levels in teh future
    if level in ['DEBUG']:
        global default_formatter
        default_formatter = "DEBUG"
        if cached_screen_handler is not None:
            _set_formatter(cached_screen_handler, 'DEBUG')
        if direct_screen_handler is not None:
            _set_formatter(direct_screen_handler, 'DEBUG')
        for l in _allLoggers:
            for this_handler in _allLoggers[l].handlers:
                _set_formatter(this_handler, 'DEBUG')


def log_user_exception(logger=None, debug=False):
    """ Log a user exception based upon a given position in the code. This is used internally despite the name and is not exposed to usersi """
    buf = cStringIO.StringIO()
    traceback.print_exc(file=buf)
    banner = 10 * '-' + ' error in user/extension code ' + 10 * '-'
    if not logger:
        logger = private_logger
    logmethod = logger.warning
    if debug:
        logmethod = logger.debug
    logmethod(banner)
    for line in buf.getvalue().splitlines():
        logmethod(line)
    logmethod('-' * len(banner))


def log_unknown_exception():
    """
    This is convenince function which is used to track down what exceptions
    are being caught by ambiguous 'except' clauses.
    It should only be called from within an exception handler.
    """
    tb_logger = getLogger('Ganga.Utility.logging.log_unknown_exception')

    # Fetch the place from where this function was called to locate the bare
    # except
    from inspect import getframeinfo, stack
    caller = getframeinfo(stack()[1][0])

    tb_logger.debug('Bare except clause triggered {0}:{1}'.format(caller.filename, caller.lineno))
    tb_logger.debug('Exception caught:', exc_info=True)




# dirac job created by ganga
from DIRAC.Core.Base.Script import parseCommandLine
parseCommandLine()
###DIRAC_IMPORT###
###DIRAC_JOB_IMPORT###
dirac = ###DIRAC_OBJECT###
j = ###JOB_OBJECT###

# default commands added by ganga
j.setName('###NAME###')
j.setExecutable('###EXE###','###EXE_ARG_STR###','###EXE_LOG_FILE###')
j.setExecutionEnv(###ENVIRONMENT###)
j.setInputSandbox(###INPUT_SANDBOX###)
j.setOutputSandbox(###OUTPUT_SANDBOX###)
j.setInputData(###INPUTDATA###)
j.setParametricInputData(###PARAMETRIC_INPUTDATA###)

###OUTPUTFILESSCRIPT###

# <-- user settings
###SETTINGS###
# user settings -->

# diracOpts added by user
###DIRAC_OPTS###

# submit the job to dirac
j.setPlatform( 'ANY' )
result = dirac.submit(j)
output(result)



from __future__ import print_function
import collections
from Ganga.Core.GangaThread.WorkerThreads.WorkerThreadPool import WorkerThreadPool
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.Utility.ColourText import getColour
from Ganga.GPIDev.Base.Proxy import getName
logger = getLogger()

_user_threadpool = None
_monitoring_threadpool = None


class ThreadPoolQueueMonitor(object):

    '''
    This class displays the user and monitor thread pools and associated queues

    The number of worker threads in the pool is initialized by
    the getConfig('Queues')['NumWorkerThreads'] config option.
    '''

    def __init__(self, user_threadpool=None, monitoring_threadpool=None):

        if user_threadpool is None:
            user_threadpool = WorkerThreadPool(worker_thread_prefix="User_Worker_")
        if monitoring_threadpool is None:
            monitoring_threadpool = WorkerThreadPool(worker_thread_prefix="Ganga_Worker_")

        global _user_threadpool
        global _monitoring_threadpool
        self._user_threadpool = _user_threadpool
        self._monitoring_threadpool = _monitoring_threadpool

        if user_threadpool != None:
            if self._user_threadpool is not None:
                self._user_threadpool.clear_queue()
                self._user_threadpool._stop_worker_threads()
                del self._user_threadpool
                del _user_threadpool
            self._user_threadpool = user_threadpool
        if monitoring_threadpool != None:
            if self._monitoring_threadpool is not None:
                self._monitoring_threadpool.clear_queue()
                self._monitoring_threadpool._stop_worker_threads()
                del self._monitoring_threadpool
                del _monitoring_threadpool
            self._monitoring_threadpool = monitoring_threadpool

        _user_threadpool = self._user_threadpool
        _monitoring_threadpool = self._monitoring_threadpool

        self._frozen = False
        self._shutdown = False

    def _display_element(self, item):
        if hasattr(item, 'name') and item.name != None:
            return item.name
        elif type(item.command_input[0]) != str:
            return getName(item.command_input[0])
        else:
            return item.command_input[0]

    def _display(self):
        '''Return the current status of the thread pools and queues.'''
        output = ''
        output += '{0:^67} | {1:^50}\n'.format('Ganga user threads:', 'Ganga monitoring threads:')
        output += '{0:^67} | {1:^50}\n'.format('------------------', '------------------------')
        output += '{0:<26} {1:<26} {2:<13} | {0:<26} {1:<28} {2:<10}\n'.format('Name', 'Command', 'Timeout')
        output += '{0:<26} {1:<26} {2:<13} | {0:<26} {1:<28} {2:<10}\n'.format('----', '-------', '-------')
        for u, m in zip(self._user_threadpool.worker_status(),
                        self._monitoring_threadpool.worker_status()):
            # name has extra spaces as colour characters are invisible but
            # still count
            if u[0] is not None:
                # User task has no ganga name
                name_user = getColour('fg.red') + str(u[0]) + getColour('fg.normal')
            else:
                # User task has a ganga name
                name_user = getColour('fg.red') + str(u[1]) + getColour('fg.normal')
            name_monitor = getColour('fg.red') + str(m[0]) + getColour('fg.normal')
            if u[1] == 'idle':
                name_user = name_user.replace(getColour('fg.red'), getColour('fg.green'))
            if m[1] == 'idle':
                name_monitor = name_monitor.replace(getColour('fg.red'), getColour('fg.green'))
            output += '{0:<35} {1:<26} {2:<15} | {3:<35} {4:<28} {5:<10}\n'.format(
                name_user, u[1][:30].replace("\n", "\\n"), u[2], name_monitor, m[1][:30].replace("\n", "\\n"), m[2])

        output += '\n'
        output += "Ganga user queue:\n"
        output += "----------------\n"
        output += str([self._display_element(elem) for elem in self._user_threadpool.get_queue()])
        output += '\n'
        output += "Ganga monitoring queue:\n"
        output += "----------------------\n"
        output += str([self._display_element(elem) for elem in self._monitoring_threadpool.get_queue()])
        return output

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('tasks...')
            return
        p.text(self._display())

    def __shouldWaitonShutdown(self):
        from Ganga.Core import getCurrentShutdownPolicy

        if getCurrentShutdownPolicy() == 'batch':
            return True
        else:
            return False

    def purge(self, force=None):
        """
        Purge the Ganga user thread pool's queue
        """
        if force is None:
            force = self.__shouldWaitonShutdown()

        _user_queue = [i for i in self._user_threadpool.get_queue()]
        queue_size = len(_user_queue)
        _actually_purge = False
        if force == True:
            _actually_purge = True
        if queue_size > 0 and not force:
            keyin = None
            while keyin == None:
                print("User queue contains unfinished tasks:")
                print(str([self._display_element(i) for i in _user_queue]))
                keyin = raw_input("Do you want to Purge the user queue ([y]/n): ")
                if keyin in ['y', '']:
                    _actually_purge = True
                elif keyin == 'n':
                    _actually_purge = False
                else:
                    print("(y/n) please!")
                    keyin = None
        if _actually_purge:
            self._user_threadpool.clear_queue()

    def _purge_all(self, force=None):
        """
        Purge ALL of the Ganga user AND Worker thread queues!
        """

        if force is None:
            force = self.__shouldWaitonShutdown()

        self.purge(force)

        _monitor_queue = [i for i in self._monitoring_threadpool.get_queue()]

        queue_size = len(_monitor_queue)
        _actually_purge = False
        if force == True:
            _actually_purge = True
        if queue_size > 0 and not force:
            keyin = None
            while keyin == None:
                print("Monitoring queue contains unfinished tasks:")
                print(str([self._display_element(i) for i in _monitor_queue]))
                keyin = raw_input("Do you want to Purge the monitoring queue ([y]/n): ")
                if keyin in ['y', '']:
                    _actually_purge = True
                elif keyin == 'n':
                    _actually_purge = False
                else:
                    print("(y/n) please")
                    keyin = None
        if _actually_purge:
            self._monitoring_threadpool.clear_queue()

    def add(self, worker_code, args=(), kwargs={}, priority=5):
        """
        Run any python callable object asynchronously through the user thread pool

        Code added to the queue will remain there until a free worker becomes
        available and picks it from the list. At this point the code will run
        asynchronously leaving the interpreter prompt unblocked and ready for user
        input.

        A simple example might be to download user data which would in the
        past block the prompt. Here I use job #72 and '/home/alex' arbitrarily
        for more info about the getOutputData command type help(Dirac.getOutputData).

        In[0]: queues.add(jobs(72).backend.getOutputData, ('/home/alex',))

        Note:
        ----

        Unlike with the queues.addProcess (where the overhead in starting up the processes 
        take a few seconds) code executes immediately when reported using monitoring
        command queues

        args:
        ----
                   worker_code = Any python callable object to run on thread 
                   args        = Any args for the callable object given here
                                 as a tuple
                   kwargs      = Any kwargs for the callable object given here
                                 as a dict
                   priority    = The thread queuing system is a priority
                                 queue with lower number = higher priority.
                                 This then should be an int normally 0-9
        """
        if not isinstance(worker_code, collections.Callable):
            logger.error('Only python callable objects can be added to the queue using queues.add()')
            logger.error('Did you perhaps try to add the return value of the function/method rather than the function/method itself')
            logger.error('e.g. Incorrect:     queues.add(myfunc()) *NOTE the brackets*')
            logger.error('e.g. Correct  :     queues.add(myfunc)')
            return
        self._user_threadpool.add_function(worker_code,
                                           args=args,
                                           kwargs=kwargs,
                                           priority=priority)

    def _addSystem(self, worker_code, args=(), kwargs={}, priority=5):

        if not isinstance(worker_code, collections.Callable):
            logger.error("Error Adding internal task!! please report this to the Ganga developers!")
            return

        if self._frozen is True:
            if not self._shutdown:
                logger.warning("Queue System is frozen not adding any more System processes!")
            return

        self._monitoring_threadpool.add_function(worker_code,
                                                 args=args,
                                                 kwargs=kwargs,
                                                 priority=priority)

    def addProcess(self,
                   command,
                   timeout=getConfig('Queues')['Timeout'],
                   env=None,
                   cwd=None,
                   shell=False,
                   eval_includes=None,
                   update_env=False,
                   priority=5,
                   callback_func=None,
                   callback_args=(),
                   callback_kwargs={},
                   fallback_func=None,
                   fallback_args=(),
                   fallback_kwargs={}):
        """
        Run a command asynchronously in a new process monitored by the user thread pool.

        Note:
        ----

        There will be no return value from this method as it runs the code asynchronously.
        stdout will go through the callback_func so if you want it displayed then use a printing function

        In[2]: def printer(x): print x
        In[0]: queues.addProcess('print 123', callback_func=printer)
        123

        Note also that a process may take a few seconds to start up unlike the thread version queues.add

        args:
        ----
                   command         = The command to run as a string 
                   timeout         = timeout for the command as int or None
                   env             = environment to run command in as dict or None
                   cwd             = working dir to run command in as string
                   shell           = True/False whether to interpret the
                                     command as a python or shell command
                   eval_includes   = This is a string that will be exec'ed before
                                     trying to eval the stdout. This allows for the
                                     the user to import certain libs before attempting
                                     to parse the output.
                   update_env      = Boolean value determining if the value of env should
                                     be updated with the environment after running the
                                     command.
                   priority        = The thread queuing system is a priority
                                     queue with lower number = higher priority.
                                     This then should be an int normally 0-9
                   callback_func   = Any python callable object. This is called
                                     once the command has finished running (or 
                                     timed out) and must take at least one arg.
                                     This first arg will be the stdout of the 
                                     executed command. This arg will be the
                                     result of unpickling stdout, falling back to 
                                     eval(stdout) such that '{}' will
                                     become a dict. Fall back to str representation
                                     if the eval fails.
                   callback_args   = Any additional args to the callback_func 
                                     are specified here as a tuple
                   callback_kwargs = kwargs for the callback_func are given here
                                     as a dict.
                   fallback_func   = Any python callable object. This is called
                                     if the command execution throws an exception.
                                     The function must take at least one arg that
                                     will be the exception that was thrown.
                   fallback_args   = Any additional args to the fallback_func 
                                     are specified here as a tuple
                   fallback_kwargs = kwargs for the fallback_func are given here
                                     as a dict.
        """
        if type(command) != type(''):
            logger.error("Input command must be of type 'string'")
            return

        if self._frozen is True:
            if not self._shutdown:
                logger.warning("Queues system is frozen. Not adding any more processes!")
            return

        self._user_threadpool.add_process(command,
                                          timeout=timeout,
                                          env=env,
                                          cwd=cwd,
                                          shell=shell,
                                          eval_includes=eval_includes,
                                          update_env=update_env,
                                          priority=priority,
                                          callback_func=callback_func,
                                          callback_args=callback_args,
                                          callback_kwargs=callback_kwargs,
                                          fallback_func=fallback_func,
                                          fallback_args=fallback_args,
                                          fallback_kwargs=fallback_kwargs)

    def threadStatus(self):
        statuses = []
        for t in self._user_threadpool.worker_status():
            if t[1] != "idle":
                statuses.append(t[0])
        for t in self._monitoring_threadpool.worker_status():
            if t[1] != "idle":
                statuses.append(t[0])
        return statuses

    def totalNumUserThreads(self):
        """Return the total number of user threads, both running and queued"""
        num = 0
        for t in self._user_threadpool.worker_status():
            if t[1] != "idle":
                num += 1

        return num + len(self._user_threadpool.get_queue())

    def totalNumIntThreads(self):
        num = 0
        for t in self._monitoring_threadpool.worker_status():
            if t[1] != "idle":
                num += 1

        return num + len(self._user_threadpool.get_queue())

    def totalNumAllThreads(self):
        """Return the total number of ALL user and worker threads currently running and queued"""
        num = 0
        num = num + self.totalNumUserThreads()
        num = num + self.totalNumIntThreads()
        return num

    def freeze(self):
        self._frozen = True
        self._user_threadpool._frozen = True
        self._monitoring_threadpool._frozen = True

    def unfreeze(self):
        self._frozen = False
        self._user_threadpool._frozen = False
        self._monitoring_threadpool._frozen = False

    def _stop_all_threads(self, shutdown=False):
        self._shutdown = shutdown
        self._user_threadpool._stop_worker_threads(shutdown)
        self._monitoring_threadpool._stop_worker_threads(shutdown)
        return

    def _start_all_threads(self):
        self._user_threadpool._start_worker_threads()
        self._monitoring_threadpool._start_worker_threads()
        return




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
'''Uses gaudirun.py to parse the job options file to allow for easy extraction
of inputdata, outputdata and output files.'''

import tempfile
import fnmatch
from Ganga.GPIDev.Base.Proxy import stripProxy, getName
from Ganga.GPIDev.Lib.File import FileBuffer
from Ganga.GPIDev.Lib.File import LocalFile
import Ganga.Utility.logging
from Ganga.Utility.util import unique
import Ganga.Utility.Config
from GangaLHCb.Lib.LHCbDataset import LHCbDataset
from Ganga.Core import ApplicationConfigurationError
from Ganga.Utility.files import expandfilename
from GangaGaudi.Lib.Applications.GaudiUtils import shellEnv_cmd
from Ganga.GPIDev.Lib.File.OutputFileManager import outputFilePostProcessingOnWN
logger = Ganga.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

## Due to a bug in Gaudi at some point we need this equivalenc here: see #204
DataObjectDescriptorCollection = str

class PythonOptionsParser:


    """ Parses job options file(s) w/ gaudirun.py to extract user's files"""

    def __init__(self, optsfiles, extraopts, env):
        self.optsfiles = optsfiles
        self.extraopts = extraopts
        self.env = env
        self.opts_dict, self.opts_pkl_str = self._get_opts_dict_and_pkl_string()

    def _get_opts_dict_and_pkl_string(self):
        '''Parse the options using gaudirun.py and create a dictionary of the
        configuration and pickle the options. The app handler will make a copy
        of the .pkl file for each job.'''
        tmp_pkl = tempfile.NamedTemporaryFile(suffix='.pkl')
        tmp_py = tempfile.NamedTemporaryFile(suffix='.py')
        py_opts = tempfile.NamedTemporaryFile(suffix='.py')
        py_opts.write(self._join_opts_files())
        py_opts.flush()

        gaudirun = 'gaudirun.py -n -v -o %s %s' \
                   % (tmp_py.name, py_opts.name)
        opts_str = ''
        err_msg = ''
        options = {}

        rc, stdout, m = shellEnv_cmd(gaudirun, self.env)

        if stdout.find('Gaudi.py') >= 0:
            msg = 'The version of gaudirun.py required for your application is not supported.'
            raise ValueError(None, msg)

        elif stdout.find('no such option: -o') >= 0:
            gaudirun = 'gaudirun.py -n -v -p %s %s' % (tmp_pkl.name, py_opts.name)
            rc, stdout, m = shellEnv_cmd(gaudirun, self.env)
            rc = 0

            if stdout and rc == 0:
                opts_str = stdout
                err_msg = 'Please check %s -v %s' % (cmdbase, py_opts.name)
                err_msg += ' returns valid python syntax'

        else:
            cmd = 'gaudirun.py -n -p %s %s' % (tmp_pkl.name, py_opts.name)
            rc, stdout, m = shellEnv_cmd(cmd, self.env)
            if rc == 0 and stdout:
                opts_str = tmp_py.read()
                err_msg = 'Please check gaudirun.py -o file.py produces a ' \
                          'valid python file.'

        if stdout and rc == 0:
            try:
                options = eval(opts_str)
            except Exception as err:
                logger.error('Cannot eval() the options file. Exception: %s', err)
                from traceback import print_exc
                logger.error(' ', print_exc())
                raise ApplicationConfigurationError(None, stdout + '###SPLIT###' + m)
            try:
                opts_pkl_string = tmp_pkl.read()
            except IOError as err:
                logger.error('Cannot read() the temporary pickle file: %s', tmp_pkl.name)
                raise err

        if not rc == 0:
            logger.debug('Failed to run: %s', gaudirun)
            raise ApplicationConfigurationError(None, stdout + '###SPLIT###' + m)

        tmp_pkl.close()
        py_opts.close()
        tmp_py.close()
        return (options, opts_pkl_string)

    def _join_opts_files(self):
        '''Create a single options file from all supplied options.'''
        joined_py_opts = ''
        for name in self.optsfiles:
            try:
                this_file = open(expandfilename(name), 'r')
                import os.path
                if os.path.splitext(name)[1] == '.py':
                    joined_py_opts += this_file.read()
                elif os.path.splitext(name)[1] == '.opts':
                    joined_py_opts += 'from Gaudi.Configuration import *\n'
                    joined_py_opts += 'importOptions(\'' + name + '\')\n'
                else:
                    msg = 'Only extensions of type ".opts" and ".py" allowed'
                    raise TypeError(msg)
            except IOError as err:
                logger.error('_join_opts_files Error: %s', str(err))
                logger.error('There was an IOError with the options file: %s', name)

        if self.extraopts:
            joined_py_opts += self.extraopts

        return joined_py_opts

    def get_input_data(self):
        '''Collects the user specified input data that the job will process'''
        data = []
        try:
            opts_input = self.opts_dict['EventSelector']['Input']
            data = [f for f in opts_input]
        except KeyError as err:
            logger.debug('No inputdata has been defined in the options file.')
            logger.debug("%s" % str(err))

        from Ganga.GPIDev.Base.Filters import allComponentFilters
        file_filter = allComponentFilters['gangafiles']
        ds = LHCbDataset()
        for d in data:
            p1 = d.find('DATAFILE=') + len('DATAFILE=')
            quote = d[p1]
            p2 = d.find(quote, p1 + 1)
            f = d[p1 + 1:p2]
            this_file = file_filter(f, None)
            if this_file is None:
                this_file = LocalFile(name=f)
            ds.files.append(this_file)
        return ds

    def get_output_files(self):
        '''Collects and organizes filenames that the job outputs'''

        sbtypes = Ganga.Utility.Config.getConfig('LHCb')['outputsandbox_types']
        outsandbox = []
        outputdata = []

        if 'CounterSummarySvc' in self.opts_dict:
            if 'xmlfile' in self.opts_dict['CounterSummarySvc']:
                f = self.opts_dict['CounterSummarySvc']['xmlfile']
                if sbtypes.count('CounterSummarySvc') > 0:
                    outsandbox.append(f)
                else:
                    outputdata.append(f)

        datatypes = ['NTupleSvc', 'EvtTupleSvc']
        for this_type in datatypes:
            if this_type in self.opts_dict:
                if 'Output' in self.opts_dict[this_type]:
                    tuples = self.opts_dict[this_type]['Output']
                    # tuple output is returned as a list
                    for t in tuples:
                        f = t.split('\'')[1]
                        if sbtypes.count(this_type) > 0:
                            outsandbox.append(f)
                        else:
                            outputdata.append(f)

        if 'HistogramPersistencySvc' in self.opts_dict:
            if 'OutputFile' in self.opts_dict['HistogramPersistencySvc']:
                f = self.opts_dict['HistogramPersistencySvc']['OutputFile']
                if sbtypes.count('HistogramPersistencySvc') > 0:
                    outsandbox.append(f)
                else:
                    outputdata.append(f)

        datatypes = ['MicroDSTStream', 'DstWriter', 'GaussTape', 'DigiWriter']
        for this_type in datatypes:
            if(this_type in self.opts_dict):
                if('Output' in self.opts_dict[this_type]):
                    # get just the file name
                    this_file = self.opts_dict[this_type]['Output'].split('\'')[1]
                    if this_file.upper().startswith('PFN:'):
                        this_file = this_file[4:]
                    if sbtypes.count(this_type) > 0:
                        outsandbox.append(this_file)
                    else:
                        outputdata.append(this_file)

        return outsandbox, outputdata

    def get_output(self, job):
        '''Builds lists of output files and output data.'''

        outputdata = [f.namePattern for f in job.outputfiles if outputFilePostProcessingOnWN(job, getName(f)) ]
        outsandbox = [f.namePattern for f in job.outputfiles if not outputFilePostProcessingOnWN(job, getName(f)) ]

        # if user put any files in both, remove them from the sandbox
        for f in outsandbox:
            if outputdata.count(f) != 0:
                outsandbox.remove(f)
                msg = 'User placed the file %s in both the outputsandbox and '
                msg += 'outputdata. It will be removed from the sandbox.'
                logger.warning(msg, f)

        gaudi_outsandbox, gaudi_outputdata = self.get_output_files()

        # handle (as best we can) any user supplied wildcards
        datalist = []  # files in sandbox that match pattern in data
        for f in outputdata:
            datalist += fnmatch.filter(gaudi_outsandbox, f)

        sandlist = []  # files in data that match sandbox pattern
        for f in outsandbox:
            sandlist += fnmatch.filter(gaudi_outputdata, f)

        datadatalist = []  # files in data that match patterns in data
        for f in outputdata:
            datadatalist += fnmatch.filter(gaudi_outputdata, f)

        # files in sandbox which match patterns in data -> data
        for f in datalist:
            gaudi_outputdata.append(f)
            if f in gaudi_outsandbox:
                gaudi_outsandbox.remove(f)
        # files in data which match patterns in sandbox but not data -> sandbox
        for f in sandlist:
            if datalist.count(f) == 0 and datadatalist.count(f) == 0:
                gaudi_outsandbox.append(f)
                if f in gaudi_outputdata:
                    gaudi_outputdata.remove(f)

        outsandbox += gaudi_outsandbox
        outputdata += gaudi_outputdata

        return unique(outsandbox), unique(outputdata)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#




#! /usr/bin/env python
###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: make_filestager_joption.py,v 1.8 2009-07-21 13:16:42 hclee Exp $
###############################################################################
# making input job option file for FileStager

import os
import os.path
import sys
import re
import pickle
import dm_util
import time

io_type = os.environ['DATASETTYPE']
io_mode = 'local'

if 'FILE_STAGER_IOMODE' in os.environ:
    io_mode = os.environ[ 'FILE_STAGER_IOMODE' ];

dq2tracertime = []
dq2tracertime.append(time.time())

## get GUID list from input_guids
guids = []
if os.path.exists('input_guids'):
    f = open('input_guids','r')
    for l in f.readlines():
        guids.append(l.strip())

## get dataset locations from os.environ['DATASETLOCATION']:
ds_locations = []
if 'DATASETLOCATION' in os.environ:
    ds_locations = os.environ['DATASETLOCATION'].split(':')

if (io_type in ['FILE_STAGER']):
    # determin site domain
    domain_replacements = {'grika.de': 'fzk.de'}
    site_domain = dm_util.get_site_domain(domain_replacements)
    print >> sys.stdout, 'detected site domain: %s' % site_domain

    # determin close se and its supported transfer protocols
    se_replacements = {'srm.cern.ch': 'srm-atlas.cern.ch'}
    close_se = dm_util.get_se_hostname(se_replacements)
    print >> sys.stdout, 'detected closed SE: %s' % close_se

    # resolve the dq2_local_site_id taking into account
    #  - file locations
    #  - site_domain
    #  - se_hostname
    #  - site configurations (for some special sites)

    # define a list of possible site ids for combined sara-nikhef site
    # allowing jobs at NIKHEF to read data from SARA and vice versa
    nl_sites = dm_util.get_srmv2_sites(cloud='NL')
    sara_nikhef_combine = []
    for s in nl_sites:
        if ( s.find('SARA-MATRIX') >= 0 ) or ( s.find('NIKHEF-ELPROD') >= 0 ):
            sara_nikhef_combine.append(s)

    force_siteid_domain = {'.*nikhef.nl.*': sara_nikhef_combine,
                           '.*sara.nl.*'  : sara_nikhef_combine}

    force_siteid_se = {'srm-disk.pic.es': ['PIC_DATADISK']}

    dq2_site_id = dm_util.resolve_dq2_local_site_id(ds_locations, site_domain, close_se, \
                                force_siteid_domain=force_siteid_domain, \
                                force_siteid_se=force_siteid_se)

    if not dq2_site_id and 'DQ2_LOCAL_SITE_ID' in os.environ:
        dq2_site_id = os.environ['DQ2_LOCAL_SITE_ID']

    print >> sys.stdout, 'detected DQ2_LOCAL_SITE_ID: %s' % dq2_site_id

    # get LFC_HOST associated with the dq2_site_id
    lfc_host = dm_util.get_lfc_host(dq2_site_id)
    print >> sys.stdout, 'LFC_HOST: %s' % lfc_host

    dq2tracertime.append(time.time())
    # resolve PFNs given the LFC_HOST and a list of GUIDs
    pfns,csum = dm_util.get_pfns(lfc_host, guids)
    dq2tracertime.append(time.time())

    #for guid in csum.keys():
    #    print >> sys.stdout, '%s %s:%s' % (guid, csum[guid]['csumtype'], csum[guid]['csumvalue'])

    # count only the PFNs on local site by match srm_endpoint of the dq2 site
    srm_endpt_info  = dm_util.get_srm_endpoint(dq2_site_id)
    print >> sys.stdout, str(srm_endpt_info)

    # define the default gridcopy protocol
    my_protocol = 'lcgcp'

    # check if DQ2_LOCAL_PROTOCOL is given: use the given protocol forcely
    if 'DQ2_LOCAL_PROTOCOL' in os.environ and os.environ['DQ2_LOCAL_PROTOCOL']:
        my_protocol = os.environ['DQ2_LOCAL_PROTOCOL']

        # just to convert the protocol name into FileStager convention
        if my_protocol in ['lcg-cp']:
            my_protocol = 'lcgcp'

    # try to detect the available protocol if io_mode is 'local'
    elif io_mode in [ 'local' ]:
        # determin the supported transfer protocols of the given SE
        protocols = dm_util.get_transfer_protocols(srm_endpt_info['se_host'])
        print >> sys.stdout, 'detected transfer protocols: %s' % repr(protocols)

        # chose a suitable protocol
        # 1. firstly remove gsiftp protocol (do we support it as an alternative of lcgcp?)
        # 2. pick up the first available protocol
        # 3. if no first protocol, use lcgcp instead
        try:
            protocols.remove('gsiftp')
        except ValueError:
            pass

        if protocols:

            ## make sure the 'file' protocl will be used if it's presented in the protocol list
            ##  - usually a case of Storm-based SE
            if 'file' in protocols:
                my_protocol = 'file'
            else:
                my_protocol = protocols[0]

    ## avoid using dcap or gsidcap until the load on SE with dcap is solved 
    #if my_protocol in ['dcap', 'gsidcap']:
    #    my_protocol = 'lcgcp'

    print >> sys.stdout, 'picked transfer protocol: %s' % my_protocol

    re_endpt = re.compile('^.*%s.*%s.*\s*$' % (srm_endpt_info['se_host'], srm_endpt_info['se_path']) )
    pfn_list = []
    pfn_csum = {}
    for guid in pfns.keys():
        print >> sys.stdout, 'guid:%s pfns:%s' % ( guid, repr(pfns[guid]) )
        for pfn in pfns[guid]:
            if re_endpt.match(pfn):
                pfn_list.append(pfn)
                pfn_csum[pfn] = csum[guid]

    # print out the PFNs and the checksum info. in LFC
    for pfn in pfn_csum.keys():
        print >> sys.stdout, '%s %s:%s' % (pfn, pfn_csum[pfn]['csumtype'], pfn_csum[pfn]['csumvalue'])
    # create a checksum list in pickle format
    fcsum = open('lfc_checksum.pickle','w')
    pickle.dump(pfn_csum, fcsum)
    fcsum.close()
    print >> sys.stdout, 'LFC checksum pickle: %s' % os.path.join(os.getcwd(), 'lfc_checksum.pickle')

    try:
        if 'ATHENA_MAX_EVENTS' in os.environ:
            evtmax = int(os.environ['ATHENA_MAX_EVENTS'])
        else:
            evtmax = -1
    except:
        evtmax = -1

    
    try:
        if 'ATHENA_SKIP_EVENTS' in os.environ:
                skipevt = int(os.environ['ATHENA_SKIP_EVENTS'])
        else:
                skipevt = 0
    except:
        skipevt = 0

    # produce the job option file for Athena/FileStager module
    fs_cp_cmd  = os.path.join( os.getcwd(), 'fs-copy.py' )
    fs_cp_args = []
    fs_of_prefix = 'file:'

    ic_jo_path = os.path.join(os.getcwd(), 'input.py')
    fs_jo_path = os.path.join(os.getcwd(), 'FileStager_jobOption.py')

    # sort pfn_list
    pfn_list.sort()

    dm_util.make_FileStager_jobOption(pfn_list, gridcopy=True, protocol=my_protocol, maxEvent=evtmax, skipEvent=skipevt, fs_cp_cmd=fs_cp_cmd, fs_cp_args=fs_cp_args, fs_of_prefix=fs_of_prefix, ic_jo_path=ic_jo_path, fs_jo_path=fs_jo_path)

    dq2tracertime.append(time.time())
    outFile = open('dq2tracertimes.txt','w')
    for itime in dq2tracertime:
        outFile.write('%s\n' % itime)
    outFile.close()


else:
    print >> sys.stderr, "make_filestager_joption.py supports only FILE_STAGER datasettype"



import os
import base64
import subprocess
import threading
import pickle
import signal
from Ganga.Core.exceptions import GangaException
from Ganga.Utility.logging import getLogger
logger = getLogger()

def env_update_script(indent=''):
    fdread, fdwrite = os.pipe()
    this_script = '''
import os, pickle
os.close(###FD_READ###)
with os.fdopen(###FD_WRITE###,'wb') as envpipe:
    pickle.dump(os.environ, envpipe)
'''
    from Ganga.GPIDev.Lib.File.FileUtils import indentScript
    script = indentScript(this_script, '###INDENT###')

    script =  script.replace('###INDENT###'  , indent      )\
                    .replace('###FD_READ###' , str(fdread) )\
                    .replace('###FD_WRITE###', str(fdwrite))
    return script, fdread, fdwrite

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def python_wrapper(command, python_setup='', update_env=False, indent=''):
    fdread, fdwrite = os.pipe()
    this_script = '''
from __future__ import print_function
import os, sys, pickle, traceback
os.close(###PKL_FDREAD###)
with os.fdopen(###PKL_FDWRITE###, 'wb') as PICKLE_STREAM:
    def output(data):
        print(pickle.dumps(data), file=PICKLE_STREAM)
    local_ns = {'pickle'        : pickle,
                'PICKLE_STREAM' : PICKLE_STREAM,
                'output'        : output}
    try:
        full_command = """###SETUP### """
        full_command += """ \n###COMMAND### """
        exec(full_command, local_ns)
    except:
        print(pickle.dumps(traceback.format_exc()), file=PICKLE_STREAM)
'''
    from Ganga.GPIDev.Lib.File.FileUtils import indentScript
    script = indentScript(this_script, '###INDENT###')

    script =  script.replace('###INDENT###'     , indent              )\
                    .replace('###SETUP###'      , python_setup.strip())\
                    .replace('###COMMAND###'    , command.strip()     )\
                    .replace('###PKL_FDREAD###' , str(fdread)         )\
                    .replace('###PKL_FDWRITE###', str(fdwrite))
    envread = None,
    envwrite = None
    if update_env:
        update_script, envread, envwrite = env_update_script()
        script += update_script
    return script, fdread, fdwrite, envread, envwrite

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def __reader(readfd, writefd, output_ns, output_var):
    os.close(writefd)
    with os.fdopen(readfd, 'rb') as read_file:
        try:
            output_ns.update({output_var: pickle.load(read_file)})
        except Exception as err:
            logger.debug("Err: %s" % str(err))
            pass  # EOFError triggered if command killed with timeout

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def __timeout_func(process, timed_out):
    if process.returncode is None:
        timed_out.set()
        try:
            os.killpg(process.pid, signal.SIGKILL)
        except Exception as e:
            logger.error("Exception trying to kill process: %s" % e)

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def execute(command,
            timeout=None,
            env=None,
            cwd=None,
            shell=True,
            python_setup='',
            eval_includes=None,
            update_env=False,
            return_code=None):
    """
    Execute an external command.
    """

    if update_env and env is None:
        raise GangaException('Cannot update the environment if None given.')

    stream_command = 'cat<&0 | sh'
    if not shell:
        stream_command = 'python -'
        command, pkl_read, pkl_write, envread, envwrite = python_wrapper(command, python_setup, update_env)
    elif update_env:
        # note the exec gets around the problem of indent and base64 gets
        # around the \n
        command_update, envread, envwrite = env_update_script()
        command += ''';python -c ""from __future__ import print_function;import base64;exec(base64.b64decode('%s'))"''' % base64.b64encode(command_update)

    if env is None and not update_env:
        pipe = subprocess.Popen('python -c "from __future__ import print_function;import os;print(os.environ)"',
                                env=None, cwd=None, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = pipe.communicate()
        env = eval(eval(str(output))[0])

        if env:
            for k, v in env.iteritems():
                if not str(v).startswith('() {'):
                    env[k] = os.path.expandvars(v)
                # Be careful with exported bash functions!
                else:
                    this_string = str(v).split('\n')
                    final_str = ""
                    for line in this_string:
                        final_str += str(os.path.expandvars(line)).strip()
                    if not final_str.endswith(';'):
                        final_str += " ;"
                    final_str += " "

    p = subprocess.Popen(stream_command,
                         shell=True,
                         env=env,
                         cwd=cwd,
                         preexec_fn=os.setsid,
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    timed_out = threading.Event()
    timer = threading.Timer(timeout, __timeout_func,
                            args=(p, timed_out))
    timer.deamon = True
    started_threads = []
    if timeout is not None:
        timer.start()
        started_threads.append(timer)

    thread_output = {}
    if not shell:
        ti = threading.Thread(target=__reader,
                              args=(pkl_read, pkl_write,
                                    thread_output, 'pkl_output'))
        ti.deamon = True
        ti.start()
        started_threads.append(ti)

    if update_env:
        ev = threading.Thread(target=__reader,
                              args=(envread, envwrite,
                                    thread_output, 'env_output'))
        ev.deamon = True
        ev.start()
        started_threads.append(ev)

    #logger.debug("Executing Command:\n'%s'" % str(command))

    stdout, stderr = p.communicate(command)
    timer.cancel()

    for t in started_threads:
        t.join()

    if stderr != '':
        # this is still debug as using the environment from dirac default_env maked a stderr message dump out
        # even though it works
        logger.debug(stderr)

    if timed_out.isSet():
        return 'Command timed out!'

    if update_env and 'env_output' in thread_output:
        env.update(thread_output['env_output'])

    if not shell and 'pkl_output' in thread_output:
        return thread_output['pkl_output']

    try:
        stdout = pickle.loads(stdout)
    except Exception as err:
        logger.debug("Err: %s" % str(err))
        local_ns = {}
        if isinstance(eval_includes, str):
            exec(eval_includes, {}, local_ns)
        try:
            stdout = eval(stdout, {}, local_ns)
        except Exception as err2:
            logger.debug("Err2: %s" % str(err2))
            pass

    return_code = p.returncode
    return stdout




import datetime
import time
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.GPIDev.Base.Proxy import isType, getName, stripProxy
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.Core import BackendError
import os.path

import Ganga.Utility.logging

import Ganga.Utility.Config

import os

logger = Ganga.Utility.logging.getLogger()

# A trival implementation of shell command with stderr/stdout capture
# This is a self-contained function (with logging).
#
# return (exitcode,soutfile,exeflag)
# soutfile - path where the stdout/stderr is stored
# exeflag - 0 if the command failed to execute, 1 if it executed
def shell_cmd(cmd, soutfile=None, allowed_exit=[0]):

    if not soutfile:
        import tempfile
        soutfile = tempfile.mktemp()

    # FIXME: garbbing stdout is done by shell magic and probably should be
    # implemented in python directly
    cmd = "%s > %s 2>&1" % (cmd, soutfile)

    logger.debug("running shell command: %s", cmd)
    rc = os.system(cmd)

    if not rc in allowed_exit:
        logger.debug('exit status [%d] of command %s', rc, cmd)
        logger.debug('full output is in file: %s', soutfile)
        with open(soutfile) as sout_file:
            logger.debug('<first 255 bytes of output>\n%s', sout_file.read(255))
        logger.debug('<end of first 255 bytes of output>')

    m = None

    if rc != 0:
        logger.debug('non-zero [%d] exit status of command %s ', rc, cmd)
        import re
        with open(soutfile) as sout_file:
            m = re.compile(r"command not found$", re.M).search(sout_file.read())

    return rc, soutfile, m is None


class Batch(IBackend):

    """ Batch submission backend.

    It  is  assumed  that   Batch  commands  (bjobs,  bsub  etc.)  setup
    correctly. As  little assumptions as  possible are made  about the
    Batch configuration but  at certain sites it may  not work correctly
    due  to a  different  Batch setup.  Tested  with CERN  and CNAF  Batch
    installations.

    Each batch system supports an 'extraopts' field, which allows customisation
    of way the job is submitted.

    PBS:
    Take environment settings on submitting machine and export to batch job:
    backend.extraopts = "-V"

    Request minimum walltime of 24 hours and minimum memory of 2GByte:
    backend.extraopts = "-l walltime=24:00:00 mem=2gb"

    The above can be combined as:
    backend.extraopts = "-V -l walltime=24:00:00 mem=2gb" 

    LSF:
    Sends mail to you when the job is dispatched and begins execution.
    backend.extraopts = "-B"

    Assigns the Ganga job name to the batch job. The job name does not need to 
    be unique.
    backend.extraopts = "-J "+ j.name

    Run the job on a host that meets the specified resource requirements.
    A resource requirement string describes the resources a job needs.
    E.g request 2Gb of memory ans 1Gb of swap space
    backend.extraopts = '-R "mem=2048" -R "swp=1024"'

    Kill job if it has exceeded the deadline (i.e. for your presentation)
    backend.extraopts = '-t 07:14:12:59' #Killed if not finished by 14 July before 1 pm
    """
    _schema = Schema(Version(1, 0), {'queue': SimpleItem(defvalue='', doc='queue name as defomed in your local Batch installation'),
                                     'extraopts': SimpleItem(defvalue='', doc='extra options for Batch. See help(Batch) for more details'),
                                     'id': SimpleItem(defvalue='', protected=1, copyable=0, doc='Batch id of the job'),
                                     'exitcode': SimpleItem(defvalue=None, typelist=[int, None], protected=1, copyable=0, doc='Process exit code'),
                                     'status': SimpleItem(defvalue='', protected=1, hidden=1, copyable=0, doc='Batch status of the job'),
                                     'actualqueue': SimpleItem(defvalue='', protected=1, copyable=0, doc='queue name where the job was submitted.'),
                                     'actualCE': SimpleItem(defvalue='', protected=1, copyable=0, doc='hostname where the job is/was running.')
                                     })
    _category = 'backends'
    _name = 'Batch'
    _hidden = 1

    def __init__(self):
        super(Batch, self).__init__()

    def command(klass, cmd, soutfile=None, allowed_exit=None):
        if allowed_exit is None:
            allowed_exit = [0]
        rc, soutfile, ef = shell_cmd(cmd, soutfile, allowed_exit)
        if not ef:
            logger.error(
                'Problem submitting batch job. Maybe your chosen batch system is not available or you have configured it wrongly')
            with open(soutfile) as sout_file:
                logger.error(sout_file.read())
                raiseable = BackendError(klass._name, 'It seems that %s commands are not installed properly:%s' % (klass._name, sout_file.readline()))
        return rc, soutfile

    command = classmethod(command)

    def submit(self, jobconfig, master_input_sandbox):

        job = self.getJobObject()

        inw = job.getInputWorkspace()
        outw = job.getOutputWorkspace()

        #scriptpath = self.preparejob(jobconfig,inw,outw)
        scriptpath = self.preparejob(jobconfig, master_input_sandbox)

        # FIX from Angelo Carbone
        # stderr_option = '-e '+str(outw.getPath())+'stderr'
        # stdout_option = '-o '+str(outw.getPath())+'stdout'

        # FIX from Alex Richards - see Savannah #87477
        stdout_option = self.config['stdoutConfig'] % str(outw.getPath())
        stderr_option = self.config['stderrConfig'] % str(outw.getPath())

        queue_option = ''
        if self.queue:
            queue_option = '-q ' + str(self.queue)

        try:
            jobnameopt = "-" + self.config['jobnameopt']
        except Exception, err:
            logger.debug("Unknown error: %s" % str(err))
            jobnameopt = False

        if self.extraopts:
            import re
            for opt in re.compile(r'(-\w+)').findall(self.extraopts):
                if opt in ('-o', '-e', '-oo', '-eo'):
                    logger.warning("option %s is forbidden", opt)
                    return False
                if self.queue and opt == '-q':
                    logger.warning("option %s is forbidden if queue is defined ( queue = '%s')", opt, self.queue)
                    return False
                if jobnameopt and opt == jobnameopt:
                    jobnameopt = False

            queue_option = queue_option + " " + self.extraopts

        if jobnameopt and job.name != '':
            # PBS doesn't like names with spaces
            tmp_name = job.name
            if isType(self, PBS):
                tmp_name = tmp_name.replace(" ", "_")
            queue_option = queue_option + " " + \
                jobnameopt + " " + "'%s'" % (tmp_name)

        # bugfix #16646
        if self.config['shared_python_executable']:
            import sys
            script_cmd = "%s %s" % (sys.executable, scriptpath)
        else:
            script_cmd = scriptpath

        command_str = self.config['submit_str'] % (inw.getPath(), queue_option, stderr_option, stdout_option, script_cmd)
        self.command_string = command_str
        rc, soutfile = self.command(command_str)
        with open(soutfile) as sout_file:
            sout = sout_file.read()
        import re
        m = re.compile(self.config['submit_res_pattern'], re.M).search(sout)
        if m is None:
            logger.warning('could not match the output and extract the Batch job identifier!')
            logger.warning('command output \n %s ', sout)
        else:
            self.id = m.group('id')
            try:
                queue = m.group('queue')
                if self.queue != queue:
                    if self.queue:
                        logger.warning('you requested queue "%s" but the job was submitted to queue "%s"', self.queue, queue)
                        logger.warning('command output \n %s ', sout)
                    else:
                        logger.info('using default queue "%s"', queue)
                    self.actualqueue = queue
            except IndexError:
                logger.info('could not match the output and extract the Batch queue name')

        # clean up the tmp file
        if os.path.exists(soutfile):
            os.remove(soutfile)

        return rc == 0

    def resubmit(self):

        job = self.getJobObject()

        inw = job.getInputWorkspace()
        outw = job.getOutputWorkspace()

        statusfilename = outw.getPath('__jobstatus__')
        try:
            os.remove(statusfilename)
        except OSError as x:
            if x.errno != 2:
                logger.warning("OSError:" + str(x))

        scriptpath = inw.getPath('__jobscript__')
        #stderr_option = '-e '+str(outw.getPath())+'stderr'
        #stdout_option = '-o '+str(outw.getPath())+'stdout'

        # FIX from Alex Richards - see Savannah #87477
        stdout_option = self.config['stdoutConfig'] % str(outw.getPath())
        stderr_option = self.config['stderrConfig'] % str(outw.getPath())

        queue_option = ''
        if self.queue:
            queue_option = '-q ' + str(self.queue)

        try:
            jobnameopt = "-" + self.config['jobnameopt']
        except Exception as err:
            logger.debug("Err: %s" % str(err))
            jobnameopt = False

        if self.extraopts:
            import re
            for opt in re.compile(r'(-\w+)').findall(self.extraopts):
                if opt in ('-o', '-e', '-oo', '-eo'):
                    logger.warning("option %s is forbidden", opt)
                    return False
                if self.queue and opt == '-q':
                    logger.warning("option %s is forbidden if queue is defined ( queue = '%s')", opt, self.queue)
                    return False
                if jobnameopt and opt == jobnameopt:
                    jobnameopt = False

            queue_option = queue_option + " " + self.extraopts

        if jobnameopt and job.name != '':
            # PBS doesn't like names with spaces
            tmp_name = job.name
            if isType(self, PBS):
                tmp_name = tmp_name.replace(" ", "_")
            queue_option = queue_option + " " + \
                jobnameopt + " " + "'%s'" % (tmp_name)

        # bugfix #16646
        if self.config['shared_python_executable']:
            import sys
            script_cmd = "%s %s" % (sys.executable, scriptpath)
        else:
            script_cmd = scriptpath

        command_str = self.config['submit_str'] % (
            inw.getPath(), queue_option, stderr_option, stdout_option, script_cmd)
        self.command_string = command_str
        rc, soutfile = self.command(command_str)
        logger.debug('from command get rc: "%d"', rc)
        if rc == 0:
            with open(soutfile) as sout_file:
                sout = sout_file.read()
            import re
            m = re.compile(
                self.config['submit_res_pattern'], re.M).search(sout)
            if m is None:
                logger.warning('could not match the output and extract the Batch job identifier!')
                logger.warning('command output \n %s ', sout)
            else:
                self.id = m.group('id')
                try:
                    queue = m.group('queue')
                    if self.queue != queue:
                        if self.queue:
                            logger.warning('you requested queue "%s" but the job was submitted to queue "%s"', self.queue, queue)
                            logger.warning('command output \n %s ', sout)
                        else:
                            logger.info('using default queue "%s"', queue)
                        self.actualqueue = queue
                except IndexError:
                    logger.info('could not match the output and extract the Batch queue name')
        else:
            with open(soutfile) as sout_file:
                logger.warning(sout_file.read())

        return rc == 0

    def kill(self):
        rc, soutfile = self.command(self.config['kill_str'] % (self.id))

        with open(soutfile) as sout_file:
            sout = sout_file.read()
        logger.debug('while killing job %s: rc = %d', self.getJobObject().getFQID('.'), rc)
        if rc == 0:
            return True
        else:
            import re
            m = re.compile(self.config['kill_res_pattern'], re.M).search(sout)
            logger.warning('while killing job %s: %s', self.getJobObject().getFQID('.'), sout)

            return m is not None

    def getStateTime(self, status):
        """Obtains the timestamps for the 'running', 'completed', and 'failed' states.

           The __jobstatus__ file in the job's output directory is read to obtain the start and stop times of the job.
           These are converted into datetime objects and returned to the user.
        """
        j = self.getJobObject()
        end_list = ['completed', 'failed']
        d = {}
        checkstr = ''

        if status == 'running':
            checkstr = 'START:'
        elif status == 'completed':
            checkstr = 'STOP:'
        elif status == 'failed':
            checkstr = 'FAILED:'
        else:
            checkstr = ''

        if checkstr == '':
            logger.debug("In getStateTime(): checkstr == ''")
            return None

        try:
            p = os.path.join(j.outputdir, '__jobstatus__')
            logger.debug("Opening output file at: %s", p)
            f = open(p)
        except IOError:
            logger.debug('unable to open file %s', p)
            return None

        for l in f:
            if checkstr in l:
                pos = l.find(checkstr)
                timestr = l[pos + len(checkstr) + 1:pos + len(checkstr) + 25]
                try:
                    t = datetime.datetime(
                        *(time.strptime(timestr, "%a %b %d %H:%M:%S %Y")[0:6]))
                except ValueError:
                    logger.debug("Value Error in file: '%s': string does not match required format.", p)
                    return None
                return t

        f.close()
        logger.debug("Reached the end of getStateTime('%s'). Returning None.", status)
        return None

    def timedetails(self):
        """Return all available timestamps from this backend.
        """
        j = self.getJobObject()
        # check for file. if it's not there don't bother calling getSateTime
        # (twice!)
        p = os.path.join(j.outputdir, '__jobstatus__')
        if not os.path.isfile(p):
            logger.error('unable to open file %s', p)
            return None

        r = self.getStateTime('running')
        c = self.getStateTime('completed')
        d = {'START': r, 'STOP': c}

        return d

    def preparejob(self, jobconfig, master_input_sandbox):

        job = self.getJobObject()
        mon = job.getMonitoringService()
        import Ganga.Core.Sandbox as Sandbox
        from Ganga.GPIDev.Lib.File import File
        from Ganga.Core.Sandbox.WNSandbox import PYTHON_DIR
        import inspect

        fileutils = File( inspect.getsourcefile(Ganga.Utility.files), subdir=PYTHON_DIR )
        subjob_input_sandbox = job.createPackedInputSandbox(jobconfig.getSandboxFiles() + [ fileutils ] )

        appscriptpath = [jobconfig.getExeString()] + jobconfig.getArgStrings()
        sharedoutputpath = job.getOutputWorkspace().getPath()
        ## FIXME Check this isn't a GangaList
        outputpatterns = jobconfig.outputbox
        environment = jobconfig.env if not jobconfig.env is None else {}


        import inspect
        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                                       'BatchScriptTemplate.py')

        from Ganga.GPIDev.Lib.File import FileUtils
        text = FileUtils.loadScript(script_location, '')

        import Ganga.Core.Sandbox as Sandbox
        import Ganga.Utility as Utility
        from Ganga.Utility.Config import getConfig
        from Ganga.GPIDev.Lib.File.OutputFileManager import getWNCodeForOutputSandbox, getWNCodeForOutputPostprocessing, getWNCodeForDownloadingInputFiles
        jobidRepr = repr(self.getJobObject().getFQID('.'))

        replace_dict = {

        '###OUTPUTSANDBOXPOSTPROCESSING###' : getWNCodeForOutputSandbox(job, ['__syslog__'], jobidRepr),

        '###OUTPUTUPLOADSPOSTPROCESSING###' : getWNCodeForOutputPostprocessing(job, ''),

        '###DOWNLOADINPUTFILES###' : getWNCodeForDownloadingInputFiles(job, ''),

        '###INLINEMODULES###' : inspect.getsource(Sandbox.WNSandbox),
        '###INLINEHOSTNAMEFUNCTION###' : inspect.getsource(Utility.util.hostname),
        '###APPSCRIPTPATH###' : repr(appscriptpath),
        #'###SHAREDINPUTPATH###' : repr(sharedinputpath)),

        '###INPUT_SANDBOX###' : repr(subjob_input_sandbox + master_input_sandbox),
        '###SHAREDOUTPUTPATH###' : repr(sharedoutputpath),

        '###OUTPUTPATTERNS###' : repr(outputpatterns),
        '###JOBID###' : jobidRepr,
        '###ENVIRONMENT###' : repr(environment),
        '###PREEXECUTE###' : self.config['preexecute'],
        '###POSTEXECUTE###' : self.config['postexecute'],
        '###JOBIDNAME###' : self.config['jobid_name'],
        '###QUEUENAME###' : self.config['queue_name'],
        '###HEARTBEATFREQUENCE###' : self.config['heartbeat_frequency'],
        '###INPUT_DIR###' : repr(job.getStringInputDir()),

        '###GANGADIR###' : repr(getConfig('System')['GANGA_PYTHONPATH'])
        }

        for k, v in replace_dict.iteritems():
            text = text.replace(str(k), str(v))

        logger.debug('subjob input sandbox %s ', subjob_input_sandbox)
        logger.debug('master input sandbox %s ', master_input_sandbox)

        from Ganga.GPIDev.Lib.File import FileBuffer

        return job.getInputWorkspace().writefile(FileBuffer('__jobscript__', text), executable=1)

    @staticmethod
    def updateMonitoringInformation(jobs):

        import re
        repid = re.compile(r'^PID: (?P<pid>\d+)', re.M)
        requeue = re.compile(r'^QUEUE: (?P<queue>\S+)', re.M)
        reactualCE = re.compile(r'^ACTUALCE: (?P<actualCE>\S+)', re.M)
        reexit = re.compile(r'^EXITCODE: (?P<exitcode>\d+)', re.M)

        def get_last_alive(f):
            """Time since the statusfile was last touched in seconds"""
            import os.path
            import time
            talive = 0
            try:
                talive = time.time() - os.path.getmtime(f)
            except OSError as x:
                logger.debug('Problem reading status file: %s (%s)', f, str(x))

            return talive

        def get_status(f):
            """Give (pid,queue,actualCE,exit code) for job"""

            pid, queue, actualCE, exitcode = None, None, None, None

            import re
            statusfile = None
            try:
                statusfile = open(f)
                stat = statusfile.read()
            except IOError as x:
                logger.debug('Problem reading status file: %s (%s)', f, str(x))
                return pid, queue, actualCE, exitcode
            finally:
                if statusfile:
                    statusfile.close()

            mpid = repid.search(stat)
            if mpid:
                pid = int(mpid.group('pid'))

            mqueue = requeue.search(stat)
            if mqueue:
                queue = str(mqueue.group('queue'))

            mactualCE = reactualCE.search(stat)
            if mactualCE:
                actualCE = str(mactualCE.group('actualCE'))

            mexit = reexit.search(stat)
            if mexit:
                exitcode = int(mexit.group('exitcode'))

            return pid, queue, actualCE, exitcode

        from Ganga.Utility.Config import getConfig
        for j in jobs:
            stripProxy(j)._getWriteAccess()
            outw = j.getOutputWorkspace()

            statusfile = os.path.join(outw.getPath(), '__jobstatus__')
            heartbeatfile = os.path.join(outw.getPath(), '__heartbeat__')
            pid, queue, actualCE, exitcode = get_status(statusfile)

            if j.status == 'submitted':
                if pid or queue:
                    j.updateStatus('running')

                    if pid:
                        j.backend.id = pid

                    if queue and queue != j.backend.actualqueue:
                        j.backend.actualqueue = queue

                    if actualCE:
                        j.backend.actualCE = actualCE

            if j.status == 'running':
                if exitcode is not None:
                    # Job has finished
                    j.backend.exitcode = exitcode
                    if exitcode == 0:
                        j.updateStatus('completed')
                    else:
                        j.updateStatus('failed')
                else:
                    # Job is still running. Check if alive
                    time = get_last_alive(heartbeatfile)
                    config = getConfig(getName(j.backend))
                    if time > config['timeout']:
                        logger.warning(
                            'Job %s has disappeared from the batch system.', str(j.getFQID('.')))
                        j.updateStatus('failed')

#_________________________________________________________________________

class LSF(Batch):

    ''' LSF backend - submit jobs to Load Sharing Facility.'''
    _schema = Batch._schema.inherit_copy()
    _category = 'backends'
    _name = 'LSF'

    config = Ganga.Utility.Config.getConfig('LSF')

    def __init__(self):
        super(LSF, self).__init__()


#_________________________________________________________________________

class PBS(Batch):

    ''' PBS backend - submit jobs to Portable Batch System.
    '''
    _schema = Batch._schema.inherit_copy()
    _category = 'backends'
    _name = 'PBS'

    config = Ganga.Utility.Config.getConfig('PBS')

    def __init__(self):
        super(PBS, self).__init__()


#_________________________________________________________________________

class SGE(Batch):

    ''' SGE backend - submit jobs to Sun Grid Engine.
    '''
    _schema = Batch._schema.inherit_copy()
    _category = 'backends'
    _name = 'SGE'

    config = Ganga.Utility.Config.getConfig('SGE')

    def __init__(self):
        super(SGE, self).__init__()




from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from fnmatch import fnmatch
import re

regex = re.compile('[*?\[\]]')


class IGangaFile(GangaObject):

    """IGangaFile represents base class for output files, such as MassStorageFile, LCGSEFile, DiracFile, LocalFile, etc 
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(
        defvalue="", doc='pattern of the file name')})
    _category = 'gangafiles'
    _name = 'IGangaFile'
    _hidden = 1

    def __init__(self):
        super(IGangaFile, self).__init__()

    def setLocation(self):
        """
        Sets the location of output files that were uploaded from the WN
        """
        raise NotImplementedError

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        raise NotImplementedError

    def get(self):
        """
        Retrieves locally all files that were uploaded before that 
        """
        raise NotImplementedError

    def getSubFiles(self, process_wildcards=False):
        """
        Returns the sub files if wildcards are used
        """
        # should we process wildcards? Used for inputfiles
        if process_wildcards:
            self.processWildcardMatches()

        # if we have subfiles, return that
        if hasattr(self, 'subfiles'):
            return self.subfiles

        return []

    def getFilenameList(self):
        """
        Returns the filenames of all associated files through a common interface
        """
        raise NotImplementedError

    def getWNScriptDownloadCommand(self, indent):
        """
        Gets the command used to download already uploaded file
        """
        raise NotImplementedError

    def put(self):
        """
        Postprocesses (upload) output file to the desired destination from the client
        """
        raise NotImplementedError

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        raise NotImplementedError

    def processWildcardMatches(self):
        """
        If namePattern contains a wildcard, populate the subfiles property
        """
        raise NotImplementedError

    def _auto_remove(self):
        """
        Remove called when job is removed as long as config option allows
        """
        self.remove()

    def _readonly(self):
        return False

    def _list_get__match__(self, to_match):
        if isinstance(to_match, str):
            return fnmatch(self.namePattern, to_match)
        # Note: type(DiracFile) = ObjectMetaclass
        # type(ObjectMetaclass) = type
        # hence checking against a class type not an instance
        if isinstance(type(to_match), type):
            return issubclass(self.__class__, to_match)
        return to_match == self

    def execSyscmdSubprocess(self, cmd):

        import subprocess

        exitcode = -999
        mystdout = ''
        mystderr = ''

        try:
            child = subprocess.Popen(
                cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (mystdout, mystderr) = child.communicate()
            exitcode = child.returncode
        finally:
            pass

        return (exitcode, mystdout, mystderr)

    def remove(self):
        """
        Objects should implement something to overload this!
        """
        raise NotImplementedError

    def accessURL(self):
        """
        Return the URL including the protocol used to access a file on a certain storage element
        """
        raise NotImplementedError

    def hasMatchedFiles(self):
        """
        Return if this file has got valid matched files. Default implementation checks for
        subfiles and locations
        """

        # check for subfiles
        if (hasattr(self, 'subfiles') and len(self.subfiles) > 0):
            # we have subfiles so we must have actual files associated
            return True

        # check for locations
        if (hasattr(self, 'locations') and len(self.locations) > 0):
            return True

        return False

    def containsWildcards(self):
        """
        Return if the name has got wildcard characters
        """
        if regex.search(self.namePattern) != None:
            return True

        return False




###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: __init__.py,v 1.1 2008-10-04 17:42:39 karl Exp $
###############################################################################
# File: Vanseg/__init__.py
# Author: K. Harrison
# Created: 070122

"""Package initialisation file"""

__author__  = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__    = "22 February 2007"
__version__ = "1.0"

from Vanseg import Vanseg
from VansegLCG import VansegLCG
from VansegLocal import VansegLocal
from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add( "Vanseg", "Interactive", VansegLocal )
allHandlers.add( "Vanseg", "Local", VansegLocal )

allHandlers.add( "Vanseg", "LCG", VansegLCG )



"""Base reporter IAction implementation.

The BaseReporter class provides an abstract reporter implementation.
 
"""

from GangaRobot.Framework.Action import IAction
from GangaRobot.Framework import Utility
from GangaRobot.Lib.Base.Extract import Node
from GangaRobot.Lib.Base.Report import Report
from Ganga.Utility.logging import getLogger

logger = getLogger()


class BaseReporter(IAction):
    
    """Base reporter IAction implementation.
    
    An abstract action implementation providing a basis for concrete run data
    reporter implementations. It contains a chain of reporters which are
    processed in the execute() method by calling handlereport() on each reporter
    in turn.
    
    Implementing a sub-class involves doing any of the following:
    - Override the __init__() method to add additional reporters to the chain
    attribute.
    - Override the handlereport() method to report on extracted data.
    
    """

    def __init__(self):
        """Create a new base reporter.
        
        The chain attribute, a list of reporters, is initialised to [self].
        
        Sub-classes can override this constructor to set the chain attribute.
        e.g.
        self.chain = [CoreReporter(), self]

        """
        self.chain = [self]

    def execute(self, runid):
        """Invoke each of the reporters in the chain.
        
        Keyword arguments:
        runid -- A UTC ID string which identifies the run.

        A run node is loaded as an instance of Extract.Node from previously
        extracted data from the configurable BaseExtractor_XmlFile, replacing
        ${runid} with the current run id.
        e.g. ~/gangadir/robot/extract/${runid}.xml

        An empty report is created as an instance of Report.Report and passed
        with the run node to the handlereport() method of each of the reporters
        in the chain.
        
        The report is saved as TEXT and HTML to the configurable
        BaseReporter_TextFile and BaseReporterHtmlFile respectively, replacing
        ${runid} with the current run id.
        e.g. ~/gangadir/robot/report/${runid}.txt
             ~/gangadir/robot/report/${runid}.html
        
        """
        logger.info("Reporting on run '%s'.", runid)
        runnode = self._loadextract(runid)
        report = Report()
        for reporter in self.chain:
            reporter.handlereport(report, runnode)
        self._savereport(report, runid)
        logger.info('Reports saved.')
        
    def _loadextract(self, runid):
        filename = Utility.expand(self.getoption('BaseExtractor_XmlFile'), runid = runid)
        content = Utility.readfile(filename)
        return Node.fromxml(content)

    def _savereport(self, report, runid):
        #text
        textfilename = Utility.expand(self.getoption('BaseReporter_TextFile'), runid = runid)
        textcontent = str(report)
        Utility.writefile(textfilename, textcontent)
        #html
        htmlfilename = Utility.expand(self.getoption('BaseReporter_HtmlFile'), runid = runid)
        htmlcontent = report.tohtml()
        Utility.writefile(htmlfilename, htmlcontent)

    def handlereport(self, report, runnode):
        """Empty default implementation.
        
        Keyword arguments:
        report -- A Report.Report for the run.
        runnode -- An Extract.Node containing the extracted data for the run.
        
        Sub-classes can override this method to modify the report or perhaps
        email or upload it, etc.
        e.g.
        def handlereport(self, report, runnode):
            report.addline('Run id   : ' + runnode.getvalue('core.id'))

        This method is called in the context of a Ganga session.
        
        """
        pass



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: GPIexport.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
##########################################################################

""" Utility for exporting symbols to GPI.
"""
from __future__ import absolute_import


# all public GPI names will be exported here
import Ganga.GPI

from .gangadoc import adddoc

from Ganga.GPIDev.Base.Proxy import isType, addProxy, getProxyClass
from Ganga.GPIDev.Base.Objects import GangaObject
from inspect import isclass

def _addToInterface(interface, name, _object):

    if isType(_object, GangaObject):
        setattr(interface, name, addProxy(_object))
    elif isclass(_object) and issubclass(_object, GangaObject):
        setattr(interface, name, getProxyClass(_object))
    else:
        setattr(interface, name, _object)

def exportToInterface(myInterface, name, _object, doc_section, docstring=None):
    '''
    Make object available publicly as "name" in the interface module. Add automatic documentation to gangadoc system.
    "doc_section" specifies how the object should be documented.
    If docstring is specified then use it to document the object (only use for "Objects" section). Otherwise use __doc__ (via pydoc utilities).
    FIXME: if you try to export the object instance, you should import it with fully qualified path, e.g.
    import X.Y.Z
    X.Y.Z.object = object
    exportToGPI("obj",X.Y.Z.object,"Objects")

    It has been observed that doing exportToGPI("obj",object,"Objects") may not work. To be understood.
    '''

    _addToInterface(myInterface, name, _object)
    adddoc(name, getattr(myInterface, name), doc_section, docstring)

def exportToGPI(name, _object, doc_section, docstring=None):
    '''
    Make object available publicly as "name" in Ganga.GPI module. Add automatic documentation to gangadoc system.
    "doc_section" specifies how the object should be documented.
    If docstring is specified then use it to document the object (only use for "Objects" section). Otherwise use __doc__ (via pydoc utilities).
    FIXME: if you try to export the object instance, you should import it with fully qualified path, e.g.
     import X.Y.Z
     X.Y.Z.object = object
     exportToGPI("obj",X.Y.Z.object,"Objects")

    It has been observed that doing exportToGPI("obj",object,"Objects") may not work. To be understood.
    '''

    exportToInterface(Ganga.GPI, name, _object, doc_section, docstring)
    adddoc(name, getattr(Ganga.GPI, name), doc_section, docstring)

def _importFromGPI(name):
    '''
    Method to import back an object from the GPI which is NOT wrapped in a proxy!
    '''
    return stripProxy(getattr(Ganga.GPI, name))

    # print 'EXPORTED',name,object
#
#
# $Log: not supported by cvs2svn $
# Revision 1.3  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.2  2006/06/21 11:42:30  moscicki
# comments
#
# Revision 1.1  2005/08/24 15:24:11  moscicki
# added docstrings for GPI objects and an interactive ganga help system based on pydoc
#
#
#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: AfsToken.py,v 1.4 2009/03/18 18:28:15 karl Exp $
##########################################################################
#
# File: AfsToken.py
# Author: K. Harrison
# Created: 060608
#
# 06/07/2006 KH:  Changed to Ganga.Utility.Shell for shell commands
#
# 07/08/2006 KH:  Added isValid() method
#
# 08/08/2006 KH:  Modified for shell being made an ICredential property
#
# 25/08/2006 KH:  Declare AfsToken class as hidden
#
# 06/09/2006 KH:  Argument minValidity added to methods create() and renew()
#
# 25/09/2006 KH:  Changed method isValid(), so that default validity is
#                 value of self.minValidity
#
# 23/11/2006 KH:  Added "pipe" keyword to option dictionary of AfsCommand
#                 Added method to determine if credential is available
#                 with system/configuration used
#                 (requests from CLT)
#
# 06/12/2006 KH:  Changed method timeleftInHMS(), to improve parsing
#                 of string returned by "tokens" command
#
# 15/02/2007 KH:  Changed method isAvailable(), to declare credential
#                 available only if ~/gangadir is on AFS, and the
#                 tokens command is found
#
# 28/02/2007 CLT: Replaced AfsCommand.options with dictionaries
#                 init_parameters, destroy_parameters, info_parameters,
#                 each providing independent options
#                 Added AfsCommand.currentOpts dictionary, to add
#                 flexibility and assist in option construction (as opposed
#                 to direct string manipulation)
#                 Added AfsToken.buildOpts(), to consolidate the option
#                 building functionality from create(), destroy() and info()
#
# 23/05/2007 KH:  Changed method isAvailable(), to also declare credential
#                 available if workspace, defined via ['FileWorkspace']topdir
#                 is on AFS, and the tokens command is found
#
# 24/05/2007 KH:  Changed method isAvailable(), to also declare credential
#                 available if local repository is used, and its location,
#                 defined via ['DefaultJobRepository']local_root
#                 is on AFS, and the tokens command is found
#
# 11/06/2007 KH:  Changed logic for checking that tokens command is found
#                 in method isAvailable()
#
# 25/09/2007 KH:  Additional argument (middleware) added to __init__ method
#                 of AfsToken, for compatibility with multi-proxy handling
#
# 08/12/2007 KH:  Changes to take into account ICommandSet being made
#                 a component class
#
# 17/12/2007 KH:  Made changes for handling of AfsTokenCommand as
#                 component class
#
# 28/02/2008 KH:  Include year information when calculating time to token
#                 expiry in timeleftInHMS() - avoid problems with leap years
#
# 28/02/2008 KH:  In isAvailable(), assume AFS is available if cell
#                 is defined
#
# 25/06/2008 KH: Remove separate checks on location of local workspace
#                and local repository, which in Ganga 5 are under directory
#                defined via ['Configuration']gangadir
#
# 02/07/2008 KH: Update to use requiresAfsToken() function of
#                Ganga.Runtime.Repository_runtime to determine
#                whether Ganga repository is on AFS
#
# 18/03/2009 MWS: Added the 'log' option to isValid()
#
#
# 18/03/2009 MWS: Added the 'force_check' argument to timeleft()
#                 and timeleftInMHS()
#

"""Module defining class for creating, querying and renewing AFS token"""
__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "05 November 2009"
__version__ = "1.16"

import os
import time

from Ganga.GPIDev.Base.Proxy import isType
from Ganga.GPIDev.Credentials.ICredential import ICommandSet, ICredential, registerCommandSet
from Ganga.GPIDev.Schema import SimpleItem
from Ganga.Runtime import Repository_runtime
from Ganga.Utility.logging import getLogger

logger = getLogger()


class AfsCommand(ICommandSet):

    """
    Class used to define shell commands and options for working with AFS token
    """

    _schema = ICommandSet._schema.inherit_copy()
    _schema['init']._meta['defvalue'] = "kinit"
    _schema['info']._meta['defvalue'] = "tokens"
    _schema['destroy']._meta['defvalue'] = "unlog"
    _schema['init_parameters']._meta['defvalue'] = {"pipe": "-pipe", "valid": "-l",
                                                    "username": "-principal", "cell": "-cell"}
    _schema['destroy_parameters']._meta['defvalue'] = {"cell": "-cell"}

    _name = "AfsCommand"
    _hidden = 1
    _enable_config = 1

    def __init__(self):
        super(AfsCommand, self).__init__()
        self.currentOpts = {}
        self.infoOpts = {}
        self.destroyOpts = {}
        return

registerCommandSet(AfsCommand)


class AfsToken (ICredential):

    """
    Class for working with AFS token
    """

    _schema = ICredential._schema.inherit_copy()
    _schema.datadict["cell"] = SimpleItem(
        defvalue="", doc="AFS cell with which token is used [empty string implies local cell]")
    _schema.datadict["username"] = SimpleItem(defvalue="",
                                              doc="AFS username with which token is used [defaults to login id]")
    _name = "AfsToken"
    _hidden = 1
    _enable_config = 1

    def __init__(self, middleware=""):
        super(AfsToken, self).__init__()
        from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
        if "ICommandSet" == self.command._name or isType(self.command, GangaList):
            self.command = AfsCommand()
        if not self.username:
            if "USERNAME" in os.environ:
                self.username = os.environ["USERNAME"]
        return

    # Populate the self.command.currentOpts dictionary with
    # AfsToken specific options.
    def buildOpts(self, command, clear=True):
        if command == self.command.init:
            if clear:
                self.command.currentOpts.clear()
            if self.username:
                self.command.currentOpts[
                    self.command.init_parameters['username']] = self.username
            if self.cell:
                self.command.currentOpts[
                    self.command.init_parameters['cell']] = self.cell
            if self.validityAtCreation:
                self.command.currentOpts[
                    self.command.init_parameters['valid']] = self.validityAtCreation
        elif command == self.command.destroy:
            if clear:
                self.command.destroyOpts.clear()
            if self.cell:
                self.command.destroyOpts[
                    self.command.destroy_parameters['cell']] = self.cell
        elif command == self.command.info:
            if clear:
                self.command.infoOpts.clear()

    def create(self, validity="", maxTry=0, minValidity="", check=False):
        self.buildOpts(self.command.init)
        return ICredential.create(self, validity, maxTry, minValidity, check)

    def destroy(self, allowed_exit=[0]):
        self.buildOpts(self.command.destroy)
        return ICredential.destroy(self, allowed_exit)

    def isAvailable(self):

        if self.cell:
            available = True
        else:
            available = False

        if not available:
            available = Repository_runtime.requiresAfsToken()

        if available:
            infoCommand = self.command.info.split()[0]
            available = False
            try:
                pathList = os.environ["PATH"].split(os.pathsep)
            except KeyError:
                pathList = []
            for searchDir in pathList:
                try:
                    fileList = os.listdir(searchDir)
                except OSError:
                    fileList = []
                if infoCommand in fileList:
                    available = True
                    break
            if available:
                logger.debug("Command '%s' found in directory '%s'" %
                             (infoCommand, searchDir))
            else:
                logger.debug("Unable to find command '%s'" % infoCommand)

        if available:
            timeleft = self.timeleft()
            if not timeleft:
                available = False

        return available

    def isValid(self, validity="", log=False, force_check=False):
        return ICredential.isValid(self, validity, log, force_check)

    def location(self):
        """
        Dummy method - returns empty string
        """
        return ""

    def renew(self, validity="", maxTry=0, minValidity="", check=True):
        return ICredential.renew(self, validity, maxTry, minValidity, check)

    def timeleft(self, units="hh:mm:ss", force_check=False):
        return ICredential.timeleft(self, units, force_check=force_check)

    def timeleftInHMS(self, force_check=False):

        localTuple = time.localtime()
        status, output, message = self.shell.cmd1(self.command.info)

        timeRemaining = "00:00:00"

        if status:
            if (1 + output.lower().find("command not found")):
                logger.warning("Command '" + self.command.info + "' not found")
                logger.warning("Unable to obtain information on AFS tokens")
                timeRemaining = ""

        if timeRemaining:
            timeString = ""
            lineList = output.split("\n")
            timeRemaining = "-1"
            for line in lineList:
                if (1 + line.lower().find("tokens for")):
                    elementList = line.rstrip("]").split("[")[1].split()
                    if self.cell:
                        afsString = "".join([" afs@", self.cell, " "])
                    else:
                        afsString = "afs@"
                    if not (1 + line.find(afsString)):
                        elementList = []
                    if len(elementList) > 1:
                        elementList.append(str(localTuple[0]))
                        timeString = " ".join(elementList[1:])
                        timeTuple = time.strptime(timeString, "%b %d %H:%M %Y")
                        timeList = list(timeTuple)
                        if localTuple[1] > timeTuple[1]:
                            timeList[0] = 1 + localTuple[0]
                        timeList[8] = localTuple[8]
                        timeTuple = tuple(timeList)
                        timeLeft = int(time.mktime(timeTuple) - time.time())
                        hours = timeLeft / (60 * 60)
                        minutes = (timeLeft - hours * 60 * 60) / 60
                        seconds = timeLeft - hours * 60 * 60 - minutes * 60
                        minString = str(minutes)
                        if len(minString) < 2:
                            minString = "0" + minString
                        secString = str(seconds)
                        if len(secString) < 2:
                            secString = "0" + secString
                        timeRemaining = "%s:%s:%s" % \
                            (str(hours), minString, secString)
                if (timeRemaining != "-1"):
                    break

        return timeRemaining

    # Add documentation strings from base class
    for method in \
            [create, destroy, isAvailable, isValid, renew, timeleft, timeleftInHMS]:
        if hasattr(ICredential, method.__name__):
            baseMethod = getattr(ICredential, method.__name__)
            setattr(method, "__doc__",
                    baseMethod.__doc__.replace("credential", "AFS token"))



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IMerger.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################
from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import isType, stripProxy
from Ganga.GPIDev.Schema import Schema, Version, ComponentItem
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList

class PostProcessException(GangaException):

    def __init__(self, x=''):
        GangaException.__init__(self, x)


class IPostProcessor(GangaObject):

    """
    Abstract base class for all classes with the postprocessor category.
    """
    _schema = Schema(Version(0, 0), {})
    _category = 'postprocessor'
    _hidden = 1

    order = 0
    success = True
    failure = False

    def __init__(self):
        super(IPostProcessor, self).__init__()

    def execute(self, job, **options):
        """
        To be overidden by inherited class
        """

        raise NotImplementedError


class MultiPostProcessor(IPostProcessor):

    """
    Contains and executes many postprocessors. This is the object which is attached to a job.
    Should behave like a list to the user.
    """

    _category = 'postprocessor'
    #_exportmethods = ['__add__', '__get__', '__str__', '__getitem__', 'append', 'remove']
    _exportmethods = ['__add__', '__get__', '__getitem__', '__len__', 'append', 'remove']
    _name = 'MultiPostProcessor'
    _schema = Schema(Version(1, 0), {
        'process_objects': ComponentItem('postprocessor', defvalue=[], hidden=1, doc='A list of Processors to run', sequence=1)
    })

    def __init__(self):
        super(MultiPostProcessor, self).__init__()

    def __construct__(self, args):
        if len(args) == 1 or len(args) > 1 and isType(args, (GangaList, list)):
            if isinstance(args, list) or isType(args, GangaList):
                for process in args:
                    self.addProcess(process)
            elif isType(args, MultiPostProcessor):
                for process in stripProxy(args).process_objects:
                    self.addProcess(process)
            else:
                self.addProcess(args)

            if hasattr(self.process_objects, 'order'):
                self.process_objects = sorted(self.process_objects, key=lambda process: process.order)
        else:
            super(MultiPostProcessor, self).__construct__(args)

    def __str__(self):
        if not isType(self.process_objects, GangaObject):
            return str(self.process_objects)
        else:
            return str(GPIProxyObjectFactory(self.process_objects))

    def append(self, value):
        self.addProcess(value)
        self.process_objects = sorted(
            self.process_objects, key=lambda process: process.order)

    def remove(self, value):
        for process in self.process_objects:
            if (isType(value, type(process)) == True):
                self.process_objects.remove(process)
                break

    def __get__(self):
        return GPIProxyObjectFactory(self.process_objects)

    def __getitem__(self, i):
        return GPIProxyObjectFactory(self.process_objects[i])

    def execute(self, job, newstatus, **options):
        # run the merger objects one at a time
        process_results = []
        for p in self.process_objects:
            # stop infinite recursion
            if p is self:
                continue
            # execute all postprocessors
            process_result = p.execute(job, newstatus, **options)
            if process_result == False:
                newstatus = 'failed'
            process_results.append(process_result)
        # if one fails then we all fail
        return not False in process_results

    def addProcess(self, process_object):
        """Adds a process object to the list of processes to be done."""
        self.process_objects.append(process_object)

    def __len__(self):
        return len(self.process_objects)

    def printSummaryTree(self, level=0, verbosity_level=0, whitespace_marker='', out=None, selection='', interactive=False):
        """If this method is overridden, the following should be noted:

        level: the hierachy level we are currently at in the object tree.
        verbosity_level: How verbose the print should be. Currently this is always 0.
        whitespace_marker: If printing on multiple lines, this allows the default indentation to be replicated.
                           The first line should never use this, as the substitution is 'name = %s' % printSummaryTree()
        out: An output stream to print to. The last line of output should be printed without a newline.'
        selection: See VPrinter for an explaintion of this.
        """
        out.write(str(self.process_objects))


from Ganga.GPIDev.Base.Filters import allComponentFilters


def postprocessor_filter(value, item):
    from Ganga.GPIDev.Lib.Job.Job import Job
    from Ganga.GPIDev.Lib.Tasks.ITransform import ITransform
    from Ganga.GPIDev.Base.Objects import ObjectMetaclass

    #from Ganga.GPIDev.Base.Proxy import stripProxy

    from Ganga.GPIDev.Base.Proxy import getProxyInterface

    valid_jobtypes = [stripProxy(i)._schema.datadict['postprocessors'] for i in getProxyInterface().__dict__.values()
                      if isinstance(stripProxy(i), ObjectMetaclass)
                      and (issubclass(stripProxy(i), Job) or issubclass(stripProxy(i), ITransform))
                      and 'postprocessors' in stripProxy(i)._schema.datadict]

    # Alex modified this line to that from above to allow for arbitrary dynamic LHCbJobTemplate etc types
#    if item is Job._schema['postprocessors']:
    if item in valid_jobtypes:
        ds = MultiPostProcessor()
        if isinstance(value, list) or isType(value, GangaList):
            ds.__construct__(value)
        else:
            ds.__construct__([value])
        return ds
    else:
        raise PostProcessException(
            "j.postprocessors only takes objects of category 'postprocessor'")

allComponentFilters['postprocessor'] = postprocessor_filter




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Adapters.IPostProcessor import PostProcessException, IPostProcessor
from Ganga.GPIDev.Base.Proxy import GPIProxyObject
from Ganga.Lib.Checkers.MetaDataChecker import MetaDataChecker
from Ganga.GPIDev.Schema import ComponentItem, FileItem, Schema, SimpleItem, Version
from Ganga.Utility.Config import makeConfig, ConfigError, getConfig
from Ganga.Utility.Plugin import allPlugins
from Ganga.Utility.logging import getLogger, log_user_exception
import commands
import copy
import os
import string
logger = getLogger()


class GaudiMetaDataChecker(MetaDataChecker):

    """
    Checks the meta data of a job is within some range,
    Currently accepts 'lumi', 'inputevents', 'outputevents', 'nskipped' and 'nfiles'.

    For example do:

    mc = GaudiMetaDataChecker()

    mc.expression = 'nskipped == 0'

    j.postprocessors.append(mc)

    to fail jobs which skip some input files.

    """
    _schema = MetaDataChecker._schema.inherit_copy()
    _category = 'postprocessor'
    _name = 'GaudiMetaDataChecker'
    _exportmethods = ['check']

    def calculateResult(self, j):
        """

        """
        inputevents = None
        outputevents = None
        lumi = None
        nskipped = None
        nfiles = None
        if self.expression.find('inputevents') > -1:
            try:
                inputevents = j.metadata['events']['input']
            except Exception, err:
                logger.error("%s" % str(err))
                raise PostProcessException("The metadata value j.events['input'] was not defined")
        if self.expression.find('outputevents') > -1:
            try:
                outputevents = j.metadata['events']['output']
            except:
                raise PostProcessException("The metadata value j.events['output'] was not defined")
        if self.expression.find('lumi') > -1:
            try:
                lumi = float(j.metadata['lumi'][1:j.metadata['lumi'].find(' ')])
            except:
                raise PostProcessException("The metadata value j.lumi was not defined")
        if self.expression.find('nskipped') > -1:
            try:
                nskipped = len(j.metadata['xmlskippedfiles'])
            except:
                raise PostProcessException("The metadata value j.xmlskippedfiles was not defined")
        if self.expression.find('nfiles') > -1:
            try:
                nfiles = float(j.metadata['xmldatanumbers']['full'])
            except:
                raise PostProcessException("The metadata value j.xmldatanumbers was not defined")
        return eval(self.expression)




#!/usr/bin/env python
from __future__ import print_function
# Interactive job wrapper created by Ganga
# ###CONSTRUCT_TIME###

###WNSANDBOX_SOURCE###

import os
import sys
import time
import glob
import mimetypes

sys.path.insert( 0, '###GANGA_PYTHONPATH###' )

statfileName = os.path.join( '###OUTPUTDIR###', '__jobstatus__' )
try:
   statfile = open( statfileName, 'w' )
except IOError, x:
   print('ERROR: Unable to write status file: %s' % statfileName)
   print('ERROR: ',x)
   raise

idfileName = os.path.join( '###OUTPUTDIR###', '__id__' )
try:
   idfile = open( idfileName, 'w' )
except IOError, x:
   print('ERROR: Unable to write id file: %s' % idfileName)
   print('ERROR: ',x)
   raise
finally:
   idfile.close()

timeString = time.strftime( '%a %d %b %H:%M:%S %Y', time.gmtime( time.time() ) )
statfile.write( 'START: ' + timeString + os.linesep )

if not os.path.exists( '###WORKDIR###'):
    os.makedirs('###WORKDIR###')
os.chdir( '###WORKDIR###' )
for inFile in ###IN_BOX###:
    if mimetypes.guess_type(inFile)[1] in ['gzip', 'bzip2']:
        getPackedInputSandbox( inFile )

###WN_INPUTFILES###

###WN_INPUTDATA###

for key, value in ###JOBCONFIG_ENV###.iteritems():
   os.environ[ key ] = value

pyCommandList = [
   'import os',
   'idfileName = \'%s\'' % idfileName,
   'idfile = open( idfileName, \'a\' )',
   'idfile.write( \'PID: \' + str( os.getppid() ) )',
   'idfile.flush()',
   'idfile.close()' ]
pyCommandString = ';'.join( pyCommandList )

commandStr = '\'' + str('###EXE_STRING### ###ARG_STRING###') + '\''
#commandStr = commandStr.replace('"', '\'')

commandList = [
   'python -c "%s' % pyCommandString,
   'os.system(%s)"' % commandStr  ]
commandString = ';'.join( commandList )

result = os.system( '%s' % commandString )

###WN_POSTPROCESSING###
for patternToZip in ###PATTERNS_TO_ZIP###:
   for currentFile in glob.glob(patternToZip):
      os.system('gzip %s' % currentFile)

createOutputSandbox( ###OUTPUT_SANDBOX_PATTERNS###, None, '###OUTPUTDIR###' )

statfile.write( 'EXITCODE: ' + str( result >> 8 ) + os.linesep )
timeString = time.strftime( '%a %d %b %H:%M:%S %Y', time.gmtime( time.time() ) )
statfile.write( 'STOP: ' + timeString + os.linesep )
statfile.flush()
statfile.close()


